"""Overrides for PTK containers which only render visible lines."""

from __future__ import annotations

import logging
from functools import lru_cache, partial
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.application.current import get_app
from prompt_toolkit.data_structures import Point
from prompt_toolkit.layout import containers as ptk_containers
from prompt_toolkit.layout.containers import (
    Container,
    HorizontalAlign,
    VerticalAlign,
    WindowAlign,
    WindowRenderInfo,
)
from prompt_toolkit.layout.controls import DummyControl as PtkDummyControl
from prompt_toolkit.layout.controls import (
    FormattedTextControl,
    UIContent,
    fragment_list_width,
    to_formatted_text,
)
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.screen import _CHAR_CACHE
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.utils import get_cwidth, take_using_weights, to_str

from euporie.core.data_structures import DiInt
from euporie.core.layout.controls import DummyControl
from euporie.core.layout.screen import BoundedWritePosition

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Callable

    from prompt_toolkit.formatted_text import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        NotImplementedOrNone,
    )
    from prompt_toolkit.layout.containers import (
        AnyContainer,
        AnyDimension,
        Float,
    )
    from prompt_toolkit.layout.margins import Margin
    from prompt_toolkit.layout.mouse_handlers import MouseHandlers
    from prompt_toolkit.layout.screen import Screen, WritePosition


log = logging.getLogger(__name__)


class DimensionTuple(NamedTuple):
    """A hashable representation of a PTK :py:class:`Dimension`."""

    min: int
    max: int
    preferred: int
    weight: int = 1


@lru_cache
def distribute_dimensions(
    size: int, dimensions: tuple[DimensionTuple, ...]
) -> list[int] | None:
    """Return the heights for all rows, or None when there is not enough space."""
    if not dimensions:
        return []

    # Sum dimensions
    sum_dimensions = DimensionTuple(
        min=sum(d.min for d in dimensions),
        max=sum(d.max for d in dimensions),
        preferred=sum(d.preferred for d in dimensions),
    )

    # If there is not enough space for both.
    # Don't do anything.
    if sum_dimensions.min > size:
        return None

    # Find optimal sizes. (Start with minimal size, increase until we cover
    # the whole size.)
    sizes = [d.min for d in dimensions]

    child_generator = take_using_weights(
        items=list(range(len(dimensions))), weights=[d.weight for d in dimensions]
    )

    i = next(child_generator)

    # Increase until we meet at least the 'preferred' size.
    preferred_stop = min(size, sum_dimensions.preferred)
    preferred_dimensions = [d.preferred for d in dimensions]

    while sum(sizes) < preferred_stop:
        if sizes[i] < preferred_dimensions[i]:
            sizes[i] += 1
        i = next(child_generator)

    # Increase until we use all the available space. (or until "max")
    if not get_app().is_done:
        max_stop = min(size, sum_dimensions.max)
        max_dimensions = [d.max for d in dimensions]

        while sum(sizes) < max_stop:
            if sizes[i] < max_dimensions[i]:
                sizes[i] += 1
            i = next(child_generator)

    return sizes


@lru_cache(maxsize=None)
class DummyContainer(Container):
    """Base class for user interface layout."""

    def __init__(self, width: int = 0, height: int = 0) -> None:
        """Define width and height if any."""
        self.width = width
        self.height = height

    def reset(self) -> None:
        """Reset the state of this container (does nothing)."""

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Return a zero-width dimension."""
        return Dimension.exact(self.width)

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return a zero-height dimension."""
        return Dimension.exact(self.height)

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Write the actual content to the screen. Does nothing."""

    def get_children(self) -> list[Container]:
        """Return an empty list of child :class:`.Container` objects."""
        return []


class HSplit(ptk_containers.HSplit):
    """Several layouts, one stacked above/under the other."""

    _pad_window: Window

    def __init__(
        self,
        children: Sequence[AnyContainer],
        window_too_small: Container | None = None,
        align: VerticalAlign = VerticalAlign.JUSTIFY,
        padding: AnyDimension = 0,
        padding_char: str | None = None,
        padding_style: str = "",
        width: AnyDimension = None,
        height: AnyDimension = None,
        z_index: int | None = None,
        modal: bool = False,
        key_bindings: KeyBindingsBase | None = None,
        style: str | Callable[[], str] = "",
    ) -> None:
        """Initialize the HSplit with a cache."""
        if window_too_small is None:
            window_too_small = DummyContainer()
        super().__init__(
            children=children,
            window_too_small=window_too_small,
            align=align,
            padding=padding,
            padding_char=padding_char,
            padding_style=padding_style,
            width=width,
            height=height,
            z_index=z_index,
            modal=modal,
            key_bindings=key_bindings,
            style=style,
        )

    def _divide_heights(self, write_position: WritePosition) -> list[int] | None:
        """Calculate and cache heights for all rows."""
        width = write_position.width
        height = write_position.height
        dimensions = [c.preferred_height(width, height) for c in self._all_children]
        result = distribute_dimensions(
            height,
            tuple(
                DimensionTuple(
                    min=d.min, max=d.max, preferred=d.preferred, weight=d.weight
                )
                for d in dimensions
            ),
        )
        return result

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Render the prompt to a `Screen` instance.

        :param screen: The :class:`~prompt_toolkit.layout.screen.Screen` class
            to which the output has to be written.
        """
        assert isinstance(write_position, BoundedWritePosition)
        sizes = self._divide_heights(write_position)
        style = parent_style + " " + to_str(self.style)
        z_index = z_index if self.z_index is None else self.z_index

        if sizes is None:
            self.window_too_small.write_to_screen(
                screen, mouse_handlers, write_position, style, erase_bg, z_index
            )
        else:
            #
            ypos = write_position.ypos
            xpos = write_position.xpos
            width = write_position.width
            bbox = write_position.bbox

            # Draw child panes.
            for s, c in zip(sizes, self._all_children):
                c.write_to_screen(
                    screen,
                    mouse_handlers,
                    BoundedWritePosition(
                        xpos,
                        ypos,
                        width,
                        s,
                        bbox=DiInt(
                            top=max(0, bbox.top - (ypos - write_position.ypos)),
                            right=bbox.right,
                            bottom=max(
                                0,
                                bbox.bottom
                                - (
                                    write_position.ypos
                                    + write_position.height
                                    - ypos
                                    - s
                                ),
                            ),
                            left=bbox.left,
                        ),
                    ),
                    style,
                    erase_bg,
                    z_index,
                )
                ypos += s

            # Fill in the remaining space. This happens when a child control
            # refuses to take more space and we don't have any padding. Adding a
            # dummy child control for this (in `self._all_children`) is not
            # desired, because in some situations, it would take more space, even
            # when it's not required. This is required to apply the styling.
            remaining_height = write_position.ypos + write_position.height - ypos
            if remaining_height > 0:
                self._remaining_space_window.write_to_screen(
                    screen,
                    mouse_handlers,
                    BoundedWritePosition(
                        xpos,
                        ypos,
                        width,
                        remaining_height,
                        bbox=DiInt(
                            top=max(0, bbox.top - (ypos - write_position.ypos)),
                            right=bbox.right,
                            bottom=min(bbox.bottom, remaining_height),
                            left=bbox.left,
                        ),
                    ),
                    style,
                    erase_bg,
                    z_index,
                )

    @property
    def pad_window(self) -> Window:
        """Create a single instance of the padding window."""
        try:
            return self._pad_window
        except AttributeError:
            self._pad_window = Window(
                height=self.padding,
                char=self.padding_char,
                style=self.padding_style,
            )
            return self._pad_window

    @property
    def _all_children(self) -> list[Container]:
        """List of child objects, including padding."""

        def get() -> list[Container]:
            result: list[Container] = []
            # Padding Top.
            if self.align in (VerticalAlign.CENTER, VerticalAlign.BOTTOM):
                result.append(Window(width=Dimension(preferred=0)))
            # The children with padding.
            for child in self.children:
                result.append(child)
                result.append(self.pad_window)
            if result:
                result.pop()
            # Padding right.
            if self.align in (VerticalAlign.CENTER, VerticalAlign.TOP):
                result.append(Window(width=Dimension(preferred=0)))
            return result

        return self._children_cache.get(tuple(self.children), get)


class VSplit(ptk_containers.VSplit):
    """Several layouts, one stacked left/right of the other."""

    _pad_window: Window

    def __init__(
        self,
        children: Sequence[AnyContainer],
        window_too_small: Container | None = None,
        align: HorizontalAlign = HorizontalAlign.JUSTIFY,
        padding: AnyDimension = 0,
        padding_char: str | None = None,
        padding_style: str = "",
        width: AnyDimension = None,
        height: AnyDimension = None,
        z_index: int | None = None,
        modal: bool = False,
        key_bindings: KeyBindingsBase | None = None,
        style: str | Callable[[], str] = "",
    ) -> None:
        """Initialize the VSplit with a cache."""
        if window_too_small is None:
            window_too_small = DummyContainer()
        super().__init__(
            children=children,
            window_too_small=window_too_small,
            align=align,
            padding=padding,
            padding_char=padding_char,
            padding_style=padding_style,
            width=width,
            height=height,
            z_index=z_index,
            modal=modal,
            key_bindings=key_bindings,
            style=style,
        )

    def _divide_widths(self, width: int) -> list[int] | None:
        """Calculate and cache widths for all columns."""
        dimensions = [c.preferred_width(width) for c in self._all_children]
        result = distribute_dimensions(
            width,
            tuple(
                DimensionTuple(
                    min=d.min, max=d.max, preferred=d.preferred, weight=d.weight
                )
                for d in dimensions
            ),
        )
        return result

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Render the prompt to a `Screen` instance.

        :param screen: The :class:`~prompt_toolkit.layout.screen.Screen` class
            to which the output has to be written.
        """
        assert isinstance(write_position, BoundedWritePosition)
        if not self.children:
            return

        children = self._all_children
        sizes = self._divide_widths(write_position.width)
        style = parent_style + " " + to_str(self.style)
        z_index = z_index if self.z_index is None else self.z_index

        # If there is not enough space.
        if sizes is None:
            self.window_too_small.write_to_screen(
                screen, mouse_handlers, write_position, style, erase_bg, z_index
            )
            return

        # Calculate heights, take the largest possible, but not larger than
        # write_position.height.
        heights = [
            child.preferred_height(width, write_position.height).preferred
            for width, child in zip(sizes, children)
        ]
        height = max(write_position.height, min(write_position.height, max(heights)))

        #
        ypos = write_position.ypos
        xpos = write_position.xpos
        bbox = write_position.bbox

        # Draw all child panes.
        for s, c in zip(sizes, children):
            c.write_to_screen(
                screen,
                mouse_handlers,
                BoundedWritePosition(
                    xpos,
                    ypos,
                    s,
                    height,
                    DiInt(
                        top=bbox.top,
                        right=max(
                            bbox.right,
                            xpos + s - write_position.xpos - write_position.width,
                        ),
                        bottom=bbox.bottom,
                        left=max(0, bbox.left - write_position.xpos - xpos),
                    ),
                ),
                style,
                erase_bg,
                z_index,
            )
            xpos += s

        # Fill in the remaining space. This happens when a child control
        # refuses to take more space and we don't have any padding. Adding a
        # dummy child control for this (in `self._all_children`) is not
        # desired, because in some situations, it would take more space, even
        # when it's not required. This is required to apply the styling.
        remaining_width = write_position.xpos + write_position.width - xpos
        if remaining_width > 0:
            self._remaining_space_window.write_to_screen(
                screen,
                mouse_handlers,
                BoundedWritePosition(
                    xpos,
                    ypos,
                    remaining_width,
                    height,
                    DiInt(
                        bbox.top,
                        max(0, bbox.left - write_position.xpos - xpos),
                        bbox.bottom,
                        max(
                            bbox.right,
                            write_position.xpos + write_position.width - xpos - s,
                        ),
                    ),
                ),
                style,
                erase_bg,
                z_index,
            )

    @property
    def pad_window(self) -> Window:
        """Create a single instance of the padding window."""
        try:
            return self._pad_window
        except AttributeError:
            self._pad_window = Window(
                width=self.padding,
                char=self.padding_char,
                style=self.padding_style,
            )
            return self._pad_window

    @property
    def _all_children(self) -> list[Container]:
        """List of child objects, including padding."""

        def get() -> list[Container]:
            result: list[Container] = []

            # Padding left.
            if self.align in (HorizontalAlign.CENTER, HorizontalAlign.RIGHT):
                result.append(Window(width=Dimension(preferred=0)))
            # The children with padding.
            for child in self.children:
                result.append(child)
                result.append(self.pad_window)
            if result:
                result.pop()
            # Padding right.
            if self.align in (HorizontalAlign.CENTER, HorizontalAlign.LEFT):
                result.append(Window(width=Dimension(preferred=0)))

            return result

        return self._children_cache.get(tuple(self.children), get)


class Window(ptk_containers.Window):
    """Container that holds a control."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize `Windows`, updating the default control for empty windows."""
        super().__init__(*args, **kwargs)
        if isinstance(self.content, PtkDummyControl):
            self.content = DummyControl()

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Write window to screen."""
        # If dont_extend_width/height was given, then reduce width/height in
        # WritePosition, if the parent wanted us to paint in a bigger area.
        # (This happens if this window is bundled with another window in a
        # HSplit/VSplit, but with different size requirements.)
        if isinstance(write_position, BoundedWritePosition):
            bbox = write_position.bbox
        else:
            bbox = DiInt(0, 0, 0, 0)
        write_position = BoundedWritePosition(
            xpos=write_position.xpos,
            ypos=write_position.ypos,
            width=write_position.width,
            height=write_position.height,
            bbox=bbox,
        )

        if self.dont_extend_width():
            write_position.width = min(
                write_position.width,
                self.preferred_width(write_position.width).preferred,
            )

        if self.dont_extend_height():
            write_position.height = min(
                write_position.height,
                self.preferred_height(
                    write_position.width, write_position.height
                ).preferred,
            )

        # Draw
        z_index = z_index if self.z_index is None else self.z_index

        draw_func = partial(
            self._write_to_screen_at_index,
            screen,
            mouse_handlers,
            write_position,
            parent_style,
            erase_bg,
        )

        if z_index is None or z_index <= 0:
            # When no z_index is given, draw right away.
            draw_func()
        else:
            # Otherwise, postpone.
            screen.draw_with_z_index(z_index=z_index, draw_func=draw_func)

    def _write_to_screen_at_index(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
    ) -> None:
        assert isinstance(write_position, BoundedWritePosition)
        # Don't bother writing invisible windows.
        # (We save some time, but also avoid applying last-line styling.)
        if write_position.height <= 0 or write_position.width <= 0:
            return

        # Calculate margin sizes.
        left_margin_widths = [self._get_margin_width(m) for m in self.left_margins]
        right_margin_widths = [self._get_margin_width(m) for m in self.right_margins]
        total_margin_width = sum(left_margin_widths + right_margin_widths)

        # Render UserControl.
        ui_content = self.content.create_content(
            write_position.width - total_margin_width, write_position.height
        )
        assert isinstance(ui_content, UIContent)

        # Scroll content.
        wrap_lines = self.wrap_lines()
        self._scroll(
            ui_content, write_position.width - total_margin_width, write_position.height
        )

        # Erase background and fill with `char`.
        self._fill_bg(screen, write_position, erase_bg)

        # Resolve `align` attribute.
        align = self.align() if callable(self.align) else self.align

        # Write body
        bbox = write_position.bbox
        visible_line_to_row_col, rowcol_to_yx = self._copy_body(
            ui_content,
            screen,
            write_position,
            sum(left_margin_widths),
            write_position.width - total_margin_width - bbox.left - bbox.right,
            self.vertical_scroll,
            self.horizontal_scroll,
            wrap_lines=wrap_lines,
            highlight_lines=True,
            vertical_scroll_2=self.vertical_scroll_2,
            always_hide_cursor=self.always_hide_cursor(),
            has_focus=get_app().layout.current_control == self.content,
            align=align,
            get_line_prefix=self.get_line_prefix,
        )

        # Remember render info. (Set before generating the margins. They need this.)
        x_offset = write_position.xpos + sum(left_margin_widths)
        y_offset = write_position.ypos

        render_info = WindowRenderInfo(
            window=self,
            ui_content=ui_content,
            horizontal_scroll=self.horizontal_scroll,
            vertical_scroll=self.vertical_scroll,
            window_width=write_position.width - total_margin_width,
            window_height=write_position.height,
            configured_scroll_offsets=self.scroll_offsets,
            visible_line_to_row_col=visible_line_to_row_col,
            rowcol_to_yx=rowcol_to_yx,
            x_offset=x_offset,
            y_offset=y_offset,
            wrap_lines=wrap_lines,
        )
        self.render_info = render_info

        # Set mouse handlers.
        def mouse_handler(mouse_event: MouseEvent) -> NotImplementedOrNone:
            """Turn screen coordinates into line coordinates."""
            # Don't handle mouse events outside of the current modal part of
            # the UI.
            if self not in get_app().layout.walk_through_modal_area():
                return NotImplemented

            # Find row/col position first.
            yx_to_rowcol = {v: k for k, v in rowcol_to_yx.items()}
            y = mouse_event.position.y
            x = mouse_event.position.x

            # If clicked below the content area, look for a position in the
            # last line instead.
            max_y = write_position.ypos + len(visible_line_to_row_col) - 1
            y = min(max_y, y)
            result: NotImplementedOrNone

            while x >= 0:
                try:
                    row, col = yx_to_rowcol[y, x]
                except KeyError:
                    # Try again. (When clicking on the right side of double
                    # width characters, or on the right side of the input.)
                    x -= 1
                else:
                    # Found position, call handler of UIControl.
                    result = self.content.mouse_handler(
                        MouseEvent(
                            position=Point(x=col, y=row),
                            event_type=mouse_event.event_type,
                            button=mouse_event.button,
                            modifiers=mouse_event.modifiers,
                        )
                    )
                    break
            else:
                # nobreak.
                # (No x/y coordinate found for the content. This happens in
                # case of a DummyControl, that does not have any content.
                # Report (0,0) instead.)
                result = self.content.mouse_handler(
                    MouseEvent(
                        position=Point(x=0, y=0),
                        event_type=mouse_event.event_type,
                        button=mouse_event.button,
                        modifiers=mouse_event.modifiers,
                    )
                )

            # If it returns NotImplemented, handle it here.
            if result == NotImplemented:
                result = self._mouse_handler(mouse_event)

            return result

        mouse_handlers.set_mouse_handler_for_range(
            x_min=write_position.xpos + max(sum(left_margin_widths), bbox.left),
            x_max=write_position.xpos
            + write_position.width
            - max(total_margin_width, bbox.right),
            y_min=write_position.ypos + bbox.top,
            y_max=write_position.ypos + write_position.height - bbox.bottom,
            handler=mouse_handler,
        )

        # Render and copy margins.
        move_x = 0

        def render_margin(m: Margin, width: int) -> UIContent:
            """Render margin. Return `Screen`."""
            # Retrieve margin fragments.
            fragments = m.create_margin(render_info, width, write_position.height)

            # Turn it into a UIContent object.
            # already rendered those fragments using this size.)
            return FormattedTextControl(fragments).create_content(
                width + 1, write_position.height
            )

        for m, width in zip(self.left_margins, left_margin_widths):
            if width > 0:  # (ConditionalMargin returns a zero width. -- Don't render.)
                # Create screen for margin.
                margin_content = render_margin(m, width)

                # Copy and shift X.
                self._copy_margin(margin_content, screen, write_position, move_x, width)
                move_x += width

        move_x = write_position.width - sum(right_margin_widths)

        for m, width in zip(self.right_margins, right_margin_widths):
            # Create screen for margin.
            margin_content = render_margin(m, width)

            # Copy and shift X.
            self._copy_margin(margin_content, screen, write_position, move_x, width)
            move_x += width

        # Apply 'self.style'
        self._apply_style(screen, write_position, parent_style)

        # Additionally apply style to line with cursor if it is visible
        if ui_content.show_cursor and not self.always_hide_cursor():
            _col, _row = ui_content.cursor_position
            if cp_yx := rowcol_to_yx.get((_row, _col)):
                self._apply_style(
                    screen,
                    BoundedWritePosition(
                        xpos=write_position.xpos,
                        ypos=cp_yx[0],
                        width=write_position.width,
                        height=1,
                    ),
                    parent_style,
                )

        # Tell the screen that this user control has been painted at this
        # position.
        screen.visible_windows_to_write_positions[self] = write_position

    def _copy_body(
        self,
        ui_content: UIContent,
        new_screen: Screen,
        write_position: WritePosition,
        move_x: int,
        width: int,
        vertical_scroll: int = 0,
        horizontal_scroll: int = 0,
        wrap_lines: bool = False,
        highlight_lines: bool = False,
        vertical_scroll_2: int = 0,
        always_hide_cursor: bool = False,
        has_focus: bool = False,
        align: WindowAlign = WindowAlign.LEFT,
        get_line_prefix: Callable[[int, int], AnyFormattedText] | None = None,
    ) -> tuple[dict[int, tuple[int, int]], dict[tuple[int, int], tuple[int, int]]]:
        """Copy the UIContent into the output screen."""
        assert isinstance(write_position, BoundedWritePosition)
        xpos = write_position.xpos + move_x
        ypos = write_position.ypos
        line_count = ui_content.line_count
        new_buffer = new_screen.data_buffer
        empty_char = _CHAR_CACHE["", ""]

        bbox = write_position.bbox

        # Map visible line number to (row, col) of input.
        # 'col' will always be zero if line wrapping is off.
        visible_line_to_row_col: dict[int, tuple[int, int]] = {}

        # Maps (row, col) from the input to (y, x) screen coordinates.
        rowcol_to_yx: dict[tuple[int, int], tuple[int, int]] = {}

        def copy_line(
            line: StyleAndTextTuples,
            lineno: int,
            x: int,
            y: int,
            is_input: bool = False,
        ) -> tuple[int, int]:
            """Copy over a single line to the output screen."""
            current_rowcol_to_yx = (
                rowcol_to_yx if is_input else {}
            )  # Throwaway dictionary.

            # Draw line prefix.
            if is_input and get_line_prefix:
                prompt = to_formatted_text(get_line_prefix(lineno, 0))
                x, y = copy_line(prompt, lineno, x, y, is_input=False)

            # Scroll horizontally.
            skipped = 0  # Characters skipped because of horizontal scrolling.
            if horizontal_scroll and is_input:
                h_scroll = horizontal_scroll
                line = explode_text_fragments(line)
                while h_scroll > 0 and line:
                    h_scroll -= get_cwidth(line[0][1])
                    skipped += 1
                    del line[:1]  # Remove first character.

                x -= h_scroll  # When scrolling over double width character,
                # this can end up being negative.

            # Align this line. (Note that this doesn't work well when we use
            # get_line_prefix and that function returns variable width prefixes.)
            if align == WindowAlign.CENTER:
                line_width = fragment_list_width(line)
                if line_width < width:
                    x += (width - line_width) // 2
            elif align == WindowAlign.RIGHT:
                line_width = fragment_list_width(line)
                if line_width < width:
                    x += width - line_width

            col = 0
            wrap_count = 0
            for style, text, *_ in line:
                new_buffer_row = new_buffer[y + ypos]

                # Remember raw VT escape sequences. (E.g. FinalTerm's
                # escape sequences.)
                if "[ZeroWidthEscape]" in style:
                    new_screen.zero_width_escapes[y + ypos][x + xpos] += text
                    continue

                for c in text:
                    char = _CHAR_CACHE[c, style]
                    char_width = char.width

                    # Wrap when the line width is exceeded.
                    if wrap_lines and x + char_width > width:
                        visible_line_to_row_col[y + 1] = (
                            lineno,
                            visible_line_to_row_col[y][1] + x,
                        )
                        y += 1
                        wrap_count += 1
                        x = 0

                        # Insert line prefix (continuation prompt).
                        if is_input and get_line_prefix:
                            prompt = to_formatted_text(
                                get_line_prefix(lineno, wrap_count)
                            )
                            x, y = copy_line(prompt, lineno, x, y, is_input=False)

                        new_buffer_row = new_buffer[y + ypos]

                        if y >= write_position.height:
                            return x, y  # Break out of all for loops.

                    # Set character in screen and shift 'x'.
                    if x >= 0 and y >= 0 and x < width:
                        new_buffer_row[x + xpos] = char

                        # When we print a multi width character, make sure
                        # to erase the neighbors positions in the screen.
                        # (The empty string if different from everything,
                        # so next redraw this cell will repaint anyway.)
                        if char_width > 1:
                            for i in range(1, char_width):
                                new_buffer_row[x + xpos + i] = empty_char

                        # If this is a zero width characters, then it's
                        # probably part of a decomposed unicode character.
                        # See: https://en.wikipedia.org/wiki/Unicode_equivalence
                        # Merge it in the previous cell.
                        elif char_width == 0:
                            # Handle all character widths. If the previous
                            # character is a multiwidth character, then
                            # merge it two positions back.
                            for pw in [2, 1]:  # Previous character width.
                                if (
                                    x - pw >= 0
                                    and new_buffer_row[x + xpos - pw].width == pw
                                ):
                                    prev_char = new_buffer_row[x + xpos - pw]
                                    char2 = _CHAR_CACHE[
                                        prev_char.char + c, prev_char.style
                                    ]
                                    new_buffer_row[x + xpos - pw] = char2

                        # Keep track of write position for each character.
                        current_rowcol_to_yx[lineno, col + skipped] = (
                            y + ypos,
                            x + xpos,
                        )

                    col += 1
                    x += char_width
            return x, y

        # Copy content.
        y = -vertical_scroll_2
        lineno = vertical_scroll

        # Render lines down to the end of the visible region (or to the cursor
        # position, whichever is lower)
        cursor_visible = ui_content.show_cursor and not self.always_hide_cursor()
        while (
            y < write_position.height - bbox.bottom
            or (cursor_visible and lineno <= ui_content.cursor_position.y)
        ) and lineno < line_count:
            visible_line_to_row_col[y] = (lineno, horizontal_scroll)

            # If lines are wrapped, we need to render all of them so we know how many
            # rows each line occupies.
            # Otherwise, we can skip rendering lines which are not visible.
            # Also always render the line with the visible cursor so we know it's position
            if (wrap_lines or bbox.top <= y) or (
                cursor_visible and lineno == ui_content.cursor_position.y
            ):
                # Take the next line and copy it in the real screen.
                line = ui_content.get_line(lineno)
                # Copy margin and actual line.
                x = 0
                x, y = copy_line(line, lineno, x, y, is_input=True)

            lineno += 1
            y += 1

        def cursor_pos_to_screen_pos(row: int, col: int) -> Point:
            """Translate row/col from UIContent to real Screen coordinates."""
            try:
                y, x = rowcol_to_yx[row, col]
            except KeyError:
                # Normally this should never happen. (It is a bug, if it happens.)
                # But to be sure, return (0, 0)
                return Point(x=0, y=0)

                # raise ValueError(
                #     'Invalid position. row=%r col=%r, vertical_scroll=%r, '
                #     'horizontal_scroll=%r, height=%r' %
                #     (row, col, vertical_scroll, horizontal_scroll, write_position.height))
            else:
                return Point(x=x, y=y)

        # Set cursor and menu positions.
        if ui_content.cursor_position:
            screen_cursor_position = cursor_pos_to_screen_pos(
                ui_content.cursor_position.y, ui_content.cursor_position.x
            )

            if has_focus:
                new_screen.set_cursor_position(self, screen_cursor_position)

                if always_hide_cursor:
                    new_screen.show_cursor = False
                else:
                    new_screen.show_cursor = ui_content.show_cursor

                self._highlight_digraph(new_screen)

            if highlight_lines:
                self._highlight_cursorlines(
                    new_screen,
                    screen_cursor_position,
                    xpos,
                    ypos,
                    width,
                    write_position.height,
                )

        # Draw input characters from the input processor queue.
        if has_focus and ui_content.cursor_position:
            self._show_key_processor_key_buffer(new_screen)

        # Set menu position.
        if ui_content.menu_position:
            new_screen.set_menu_position(
                self,
                cursor_pos_to_screen_pos(
                    ui_content.menu_position.y, ui_content.menu_position.x
                ),
            )

        # Update output screen height.
        new_screen.height = max(new_screen.height, ypos + write_position.height)

        return visible_line_to_row_col, rowcol_to_yx

    def _copy_margin(
        self,
        margin_content: UIContent,
        new_screen: Screen,
        write_position: WritePosition,
        move_x: int,
        width: int,
    ) -> None:
        """Copy characters from the margin screen to the real screen."""
        assert isinstance(write_position, BoundedWritePosition)
        xpos = write_position.xpos + move_x
        ypos = write_position.ypos
        wp_bbox = write_position.bbox

        margin_write_position = BoundedWritePosition(
            xpos, ypos, width, write_position.height, wp_bbox
        )
        self._copy_body(margin_content, new_screen, margin_write_position, 0, width)

    def _fill_bg(
        self, screen: Screen, write_position: WritePosition, erase_bg: bool
    ) -> None:
        """Erase/fill the background."""
        assert isinstance(write_position, BoundedWritePosition)
        char: str | None
        char = self.char() if callable(self.char) else self.char

        if erase_bg or char:
            wp = write_position
            char_obj = _CHAR_CACHE[char or " ", ""]

            bbox = wp.bbox
            for y in range(wp.ypos + bbox.top, wp.ypos + wp.height - bbox.bottom):
                row = screen.data_buffer[y]
                for x in range(wp.xpos + bbox.left, wp.xpos + wp.width - bbox.right):
                    row[x] = char_obj

    def _apply_style(
        self,
        new_screen: Screen,
        write_position: WritePosition,
        parent_style: str,
    ) -> None:
        # Apply `self.style`.
        style = f"{parent_style} {to_str(self.style)}"

        new_screen.fill_area(write_position, style=style, after=False)

        # Apply the 'last-line' class to the last line of each Window. This can
        # be used to apply an 'underline' to the user control.
        if isinstance(write_position, BoundedWritePosition):
            if write_position.bbox.bottom == 0:
                wp = BoundedWritePosition(
                    write_position.xpos,
                    write_position.ypos + write_position.height - 1,
                    write_position.width,
                    1,
                )
                new_screen.fill_area(wp, "class:last-line", after=True)
        else:
            new_screen.fill_area(write_position, "class:last-line", after=True)

    def _mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Mouse handler. Called when the UI control doesn't handle this particular event.

        Return `NotImplemented` if nothing was done as a consequence of this
        key binding (no UI invalidate required in that case).
        """
        if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            return self._scroll_down()
        elif mouse_event.event_type == MouseEventType.SCROLL_UP:
            return self._scroll_up()

        return NotImplemented

    def _scroll_down(self) -> NotImplementedOrNone:  # type: ignore [override]
        """Scroll window down."""
        info = self.render_info

        if info is None:
            return NotImplemented

        if self.vertical_scroll < info.content_height - info.window_height:
            if info.cursor_position.y <= info.configured_scroll_offsets.top:
                self.content.move_cursor_down()
            self.vertical_scroll += 1
            return None

        return NotImplemented

    def _scroll_up(self) -> NotImplementedOrNone:  # type: ignore [override]
        """Scroll window up."""
        info = self.render_info

        if info is None:
            return NotImplemented

        if info.vertical_scroll > 0:
            # TODO: not entirely correct yet in case of line wrapping and long lines.
            if (
                info.cursor_position.y
                >= info.window_height - 1 - info.configured_scroll_offsets.bottom
            ):
                self.content.move_cursor_up()
            self.vertical_scroll -= 1
            return None

        return NotImplemented


class FloatContainer(ptk_containers.FloatContainer):
    """A `FloatContainer` which uses :py`BoundedWritePosition`s."""

    def _draw_float(
        self,
        fl: Float,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Draw a single Float."""
        # When a menu_position was given, use this instead of the cursor
        # position. (These cursor positions are absolute, translate again
        # relative to the write_position.)
        # Note: This should be inside the for-loop, because one float could
        #       set the cursor position to be used for the next one.
        cpos = screen.get_menu_position(
            fl.attach_to_window or get_app().layout.current_window
        )
        cursor_position = Point(
            x=cpos.x - write_position.xpos, y=cpos.y - write_position.ypos
        )

        fl_width = fl.get_width()
        fl_height = fl.get_height()
        width: int
        height: int
        xpos: int
        ypos: int

        # Left & width given.
        if fl.left is not None and fl_width is not None:
            xpos = fl.left
            width = fl_width
        # Left & right given -> calculate width.
        elif fl.left is not None and fl.right is not None:
            xpos = fl.left
            width = write_position.width - fl.left - fl.right
        # Width & right given -> calculate left.
        elif fl_width is not None and fl.right is not None:
            xpos = write_position.width - fl.right - fl_width
            width = fl_width
        # Near x position of cursor.
        elif fl.xcursor:
            if fl_width is None:
                width = fl.content.preferred_width(write_position.width).preferred
                width = min(write_position.width, width)
            else:
                width = fl_width

            xpos = cursor_position.x
            if xpos + width > write_position.width:
                xpos = max(0, write_position.width - width)
        # Only width given -> center horizontally.
        elif fl_width:
            xpos = int((write_position.width - fl_width) / 2)
            width = fl_width
        # Otherwise, take preferred width from float content.
        else:
            width = fl.content.preferred_width(write_position.width).preferred

            if fl.left is not None:
                xpos = fl.left
            elif fl.right is not None:
                xpos = max(0, write_position.width - width - fl.right)
            else:  # Center horizontally.
                xpos = max(0, int((write_position.width - width) / 2))

            # Trim.
            width = min(width, write_position.width - xpos)

        # Top & height given.
        if fl.top is not None and fl_height is not None:
            ypos = fl.top
            height = fl_height
        # Top & bottom given -> calculate height.
        elif fl.top is not None and fl.bottom is not None:
            ypos = fl.top
            height = write_position.height - fl.top - fl.bottom
        # Height & bottom given -> calculate top.
        elif fl_height is not None and fl.bottom is not None:
            ypos = write_position.height - fl_height - fl.bottom
            height = fl_height
        # Near cursor.
        elif fl.ycursor:
            ypos = cursor_position.y + (0 if fl.allow_cover_cursor else 1)

            if fl_height is None:
                height = fl.content.preferred_height(
                    width, write_position.height
                ).preferred
            else:
                height = fl_height

            # Reduce height if not enough space. (We can use the height
            # when the content requires it.)
            if height > write_position.height - ypos:
                if write_position.height - ypos + 1 >= ypos:
                    # When the space below the cursor is more than
                    # the space above, just reduce the height.
                    height = write_position.height - ypos
                else:
                    # Otherwise, fit the float above the cursor.
                    height = min(height, cursor_position.y)
                    ypos = cursor_position.y - height

        # Only height given -> center vertically.
        elif fl_height:
            ypos = int((write_position.height - fl_height) / 2)
            height = fl_height
        # Otherwise, take preferred height from content.
        else:
            height = fl.content.preferred_height(width, write_position.height).preferred

            if fl.top is not None:
                ypos = fl.top
            elif fl.bottom is not None:
                ypos = max(0, write_position.height - height - fl.bottom)
            else:  # Center vertically.
                ypos = max(0, int((write_position.height - height) / 2))

            # Trim.
            height = min(height, write_position.height - ypos)

        # Write float.
        # (xpos and ypos can be negative: a float can be partially visible.)
        if height > 0 and width > 0:
            wp = BoundedWritePosition(
                xpos=xpos + write_position.xpos,
                ypos=ypos + write_position.ypos,
                width=width,
                height=height,
            )

            if not fl.hide_when_covering_content or self._area_is_empty(screen, wp):
                fl.content.write_to_screen(
                    screen,
                    mouse_handlers,
                    wp,
                    style,
                    erase_bg=not fl.transparent(),
                    z_index=z_index,
                )


ptk_containers.HSplit = HSplit  # type: ignore[misc]
ptk_containers.VSplit = VSplit  # type: ignore[misc]
ptk_containers.Window = Window  # type: ignore[misc]
ptk_containers.FloatContainer = FloatContainer  # type: ignore[misc]



"""Contain functions which convert data to html format."""

from __future__ import annotations

import logging
from functools import cache
from typing import TYPE_CHECKING

from euporie.core.app.current import get_app
from euporie.core.convert.registry import register
from euporie.core.lexers import detect_lexer

if TYPE_CHECKING:
    from markdown_it import MarkdownIt

    from euporie.core.convert.datum import Datum

log = logging.getLogger(__name__)


@cache
def markdown_parser() -> MarkdownIt:
    """Lazy-load a markdown parser."""
    from markdown_it import MarkdownIt
    from mdit_py_plugins.amsmath import amsmath_plugin
    from mdit_py_plugins.dollarmath.index import dollarmath_plugin
    from mdit_py_plugins.texmath.index import texmath_plugin
    from pygments import highlight
    from pygments.formatters import HtmlFormatter

    class MarkdownParser(MarkdownIt):
        """Subclas the markdown parser to allow ``file:`` URIs."""

        def validateLink(self, url: str) -> bool:
            """Allow all link URIs."""
            return True

    return (
        (
            MarkdownParser(
                options_update={
                    "highlight": lambda text, language, lang_args: highlight(
                        text,
                        detect_lexer(text, language=language),
                        HtmlFormatter(
                            nowrap=True,
                            noclasses=True,
                            style=(
                                app.syntax_theme
                                if hasattr((app := get_app()), "syntax_theme")
                                else "default"
                            ),
                        ),
                    )
                }
            )
            .enable("linkify")
            .enable("table")
            .enable("strikethrough")
        )
        .use(texmath_plugin)
        .use(dollarmath_plugin)
        .use(amsmath_plugin)
        # .use(tasklists_plugin)
    )


@register(from_="markdown", to="html")
async def markdown_to_html_markdown_it(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str:
    """Convert markdown to HTML using :py:mod:`markdownit_py`."""
    parser = markdown_parser()
    data = datum.data
    markup = data.decode() if isinstance(data, bytes) else data
    return parser.render(markup)



"""Initiate logging for euporie.core."""

from __future__ import annotations

import logging
import logging.config
import sys
import textwrap
from collections import deque
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

from prompt_toolkit.application.current import get_app_session
from prompt_toolkit.formatted_text.base import FormattedText
from prompt_toolkit.output.defaults import create_output
from prompt_toolkit.renderer import (
    print_formatted_text as renderer_print_formatted_text,
)
from prompt_toolkit.shortcuts.utils import print_formatted_text
from prompt_toolkit.styles.pygments import style_from_pygments_cls
from prompt_toolkit.styles.style import Style, merge_styles

from euporie.core.ft.utils import indent, lex, wrap
from euporie.core.io import PseudoTTY
from euporie.core.style import LOG_STYLE, get_style_by_name
from euporie.core.utils import dict_merge

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any, Callable, TextIO

    from prompt_toolkit.formatted_text.base import StyleAndTextTuples
    from prompt_toolkit.styles.base import BaseStyle

    from euporie.core.config import Config

log = logging.getLogger(__name__)

LOG_QUEUE: deque = deque(maxlen=1000)


class BufferedLogs(logging.Handler):
    """A handler that collects log records and replays them on exit."""

    def __init__(self, logger: logging.Logger | None = None) -> None:
        """Initialize the collector.

        Args:
            logger: Logger to collect from and replay to. If None, uses root logger.
        """
        super().__init__()
        self.records: list[logging.LogRecord] = []
        self._logger = logger or logging.getLogger()
        self._original_handlers: list[logging.Handler] = []

    def emit(self, record: logging.LogRecord) -> None:
        """Store the log record."""
        self.records.append(record)

    def replay(self) -> None:
        """Replay collected logs through the original logger."""
        for record in self.records:
            if record.exc_info:
                # Create a new record to avoid issues with stale exc_info
                record = logging.LogRecord(
                    record.name,
                    record.levelno,
                    record.pathname,
                    record.lineno,
                    record.msg,
                    record.args,
                    record.exc_info,
                    record.funcName,
                )
            self._logger.handle(record)

    def __enter__(self) -> BufferedLogs:
        """Store and replace the log handlers."""
        # Save and remove existing handlers
        self._original_handlers = self._logger.handlers[:]
        self._logger.handlers.clear()
        # Add ourselves as the only handler
        self._logger.addHandler(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Restore the original handlers."""
        # Remove ourselves
        self._logger.removeHandler(self)
        # Restore original handlers
        self._logger.handlers = self._original_handlers
        # Replay collected records through original handlers
        self.replay()


class FtFormatter(logging.Formatter):
    """Base class for formatted text logging formatter."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Create a new formatter instance."""
        super().__init__(*args, **kwargs)
        self.datefmt = self.datefmt or "%H:%M:%S"

    def prepare(
        self, record: logging.LogRecord, width: int | None = None
    ) -> logging.LogRecord:
        """Format certain attributes on the log record."""
        record.asctime = self.formatTime(record, self.datefmt)
        record.message = record.getMessage()
        record.exc_text = ""
        if record.exc_info and not record.exc_text:
            record.exc_text = self.formatException(record.exc_info)
        return record

    def format_traceback(self, tb: str) -> StyleAndTextTuples:
        """Format a traceback string using pygments."""
        return lex([("", tb)], "pytb")

    def ft_format(
        self, record: logging.LogRecord, width: int | None = None
    ) -> FormattedText:
        """Format a log record as :py:class:`FormattedText`."""
        return FormattedText([])


class FormattedTextHandler(logging.StreamHandler):
    """Format log records for display on the standard output."""

    formatter: FtFormatter

    def __init__(
        self,
        stream: str | TextIO | PseudoTTY | None = None,
        share_stream: bool = True,
        style: BaseStyle | None = None,
        pygments_theme: str = "euporie",
    ) -> None:
        """Create a new log handler instance."""
        # If a filename string is passed, open it as a stream
        if isinstance(stream, str):
            # We fake a TTY so we can output color

            stream = PseudoTTY(open(stream, "a"))  # noqa: SIM115,PTH123

        super().__init__(stream)
        self.share_stream = share_stream
        self.style = style or merge_styles(
            [
                style_from_pygments_cls(get_style_by_name(pygments_theme)),
                style or Style(LOG_STYLE),
            ]
        )

        self.output = create_output(stdout=self.stream)

    def ft_format(self, record: logging.LogRecord) -> FormattedText:
        """Format the specified record."""
        if self.formatter is not None:
            return self.formatter.ft_format(record, width=self.output.get_size()[1])
        else:
            return FormattedText([])

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a formatted record."""
        try:
            msg = self.ft_format(record)
            if self.share_stream:
                print_formatted_text(
                    msg,
                    end="",
                    style=self.style,
                    output=self.output,
                    include_default_pygments_style=False,
                )
            else:
                renderer_print_formatted_text(
                    output=self.output,
                    formatted_text=msg,
                    style=self.style,
                )

        except RecursionError:
            raise
        except Exception:
            self.handleError(record)


class QueueHandler(logging.Handler):
    """This handler store logs events into a queue."""

    formatter: FtFormatter
    hook_id = 0
    hooks: ClassVar[dict[int, Callable]] = {}

    def __init__(
        self,
        *args: Any,
        queue: deque,
        style: Style | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize an instance, using the passed queue."""
        logging.Handler.__init__(self, *args, **kwargs)
        self.queue = queue
        self.style = style or Style(LOG_STYLE)

    def emit(self, record: logging.LogRecord) -> None:
        """Queue unformatted records, as they will be formatted when accessed."""
        message = self.formatter.ft_format(record)
        self.queue.append(message)
        for hook in self.hooks.values():
            if callable(hook):
                hook(message)

    @classmethod
    def hook(cls, hook: Callable) -> int:
        """Add a hook to run after each log entry.

        Args:
            hook: The hook function to add

        Returns:
            The hook id
        """
        hook_id = cls.hook_id
        cls.hook_id += 1
        cls.hooks[hook_id] = hook
        return hook_id

    @classmethod
    def unhook(cls, hook_id: int) -> None:
        """Remove a hook function.

        Args:
            hook_id: The ID of the hook function to remove
        """
        if hook_id in cls.hooks:
            del cls.hooks[hook_id]


class LogTabFormatter(FtFormatter):
    """Format log messages for display in the log view tab."""

    def ft_format(
        self, record: logging.LogRecord, width: int | None = None
    ) -> FormattedText:
        """Format a log record as formatted text."""
        record = self.prepare(record)
        output: StyleAndTextTuples = [
            ("", "["),
            ("class:pygments.literal.date", f"{record.asctime}"),
            ("", "] ["),
            (f"class:log.level.{record.levelname}", f"{record.levelname}"),
            ("", "] ["),
            ("class:pygments.comment", f"{record.name}"),
            ("", "] "),
            ("class:log,msg", record.message),
            ("", "\n"),
        ]
        if record.exc_text:
            output += [*self.format_traceback(record.exc_text), ("", "\n")]
        return FormattedText(output)


class StdoutFormatter(FtFormatter):
    """A log formatter for formatting log entries for display on the standard output."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Create a new formatter instance."""
        super().__init__(*args, **kwargs)
        self.last_date: str | None = None

    def ft_format(
        self, record: logging.LogRecord, width: int | None = None
    ) -> FormattedText:
        """Format log records for display on the standard output."""
        if width is None:
            width = get_app_session().output.get_size()[1]

        record = self.prepare(record)
        # path_name = Path(record.pathname).name

        date = f"{record.asctime}"
        if date == self.last_date:
            date = " " * len(date)
        else:
            self.last_date = date
        # ref = f"{path_name}:{record.lineno}"
        ref = f"{record.name}.{record.funcName}:{record.lineno}"

        msg_pad = len(date) + 10
        msg_pad_1st_line = msg_pad + 1 + len(ref)

        msg_lines = "\n".join(
            textwrap.wrap(
                record.message,
                width=width,
                initial_indent=" " * msg_pad_1st_line,
                replace_whitespace=False,
            )
        ).split("\n")
        subsequent_indent = " " * msg_pad
        for i in range(1, len(msg_lines)):
            msg_lines[i] = f"{subsequent_indent}{msg_lines[i]}"

        output: StyleAndTextTuples = [
            ("class:pygments.literal.date", date),
            ("", " " * (9 - len(record.levelname))),
            (f"class:log.level.{record.levelname}", record.levelname),
            ("", " "),
            (
                "class:pygments.text",
                msg_lines[0].strip().ljust(width - msg_pad_1st_line),
            ),
            ("", " "),
            ("class:pygments.comment", ref),
        ]
        for line in msg_lines[1:]:
            output += [
                ("", "\n"),
                ("class:log,msg", line),
            ]
        if record.exc_text:
            output += indent(
                wrap(
                    self.format_traceback(record.exc_text),
                    width=width - msg_pad,
                    margin="  ",
                ),
                margin=" " * msg_pad,
            )
        output += [("", "\n")]
        return FormattedText(output)


class stdout_to_log:
    """A context manager which captures standard output and logs it."""

    def __init__(
        self, log: logging.Logger, output: str = "Literal['stdout','stderr']"
    ) -> None:
        """Create a new instance of the capturing context manager.

        Args:
            log: The logger to send the output to
            output: Whether to capture the standard output or the standard error

        """
        self.log = log
        self.out = StringIO()
        self.output = output
        self._original: TextIO | None = None

    def __enter__(self) -> None:
        """Intercept the standard output when entering the context manager."""
        if self.output == "stderr":
            self._original = sys.stderr
            sys.stderr = self.out
        else:
            self._original = sys.stdout
            sys.stdout = self.out

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Replace the standard output, and logs the captured output."""
        assert self._original is not None
        if self.output == "stderr":
            sys.stderr = self._original
        else:
            sys.stdout = self._original
        if exc_type is not None:
            self.out.seek(0)
            for line in self.out.readlines():
                self.log.debug(str(line).strip())
        self.out.close()


def handle_exception(
    exc_type: type[BaseException],
    exc_value: BaseException,
    exc_traceback: TracebackType | None,
) -> Any:
    """Log unhandled exceptions and their tracebacks in the log.

    Args:
        exc_type: The type of the exception
        exc_value: The exception instance
        exc_traceback: The associated traceback
    """
    # Check the exception is not a keyboard interrupt (Ctrl+C) - if so, so not log it
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    # Log the exception
    log.critical("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


def add_log_level(name: str, number: int) -> None:
    """Add a new level to the logger."""
    level_name = name.upper()
    method_name = name.lower()

    if not hasattr(logging, level_name):

        def _log_for_level(
            self: logging.Logger, message: str, *args: Any, **kwargs: Any
        ) -> None:
            if self.isEnabledFor(number):
                self._log(number, message, args, **kwargs)

        def _log_to_root(message: str, *args: Any, **kwargs: Any) -> None:
            logging.log(number, message, *args, **kwargs)

        logging.addLevelName(number, level_name)
        setattr(logging, level_name, number)
        setattr(logging.getLoggerClass(), method_name, _log_for_level)
        setattr(logging, method_name, _log_to_root)


def setup_logs(config: Config | None = None) -> None:
    """Configure the logger for euporie."""
    # Add custom log levels
    # add_log_level("kernelio", logging.WARNING - 1)
    # add_log_level("kernel", logging.INFO - 1)
    # add_log_level("convert", logging.INFO - 2)
    # add_log_level("ui", logging.INFO - 3)

    # Default log config
    log_config: dict[str, Any] = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "file_format": {
                "format": "{asctime}.{msecs:03.0f} {levelname:<7} [{name}.{funcName}:{lineno}] {message}",
                "style": "{",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
            "stdout_format": {
                "()": StdoutFormatter,
            },
            "log_tab_format": {
                "()": LogTabFormatter,
            },
        },
        "handlers": {
            "stdout": {
                "level": "INFO",
                "()": FormattedTextHandler,
                "formatter": "stdout_format",
                "stream": sys.stdout,
            },
            "log_tab": {
                "level": "INFO",
                "()": QueueHandler,
                "formatter": "log_tab_format",
                "queue": LOG_QUEUE,
            },
        },
        "loggers": {
            "euporie": {
                "level": "INFO",
                "handlers": ["log_tab", "stdout"],
                "propagate": False,
            },
        },
        # Log everything to the internal logger
        "root": {"handlers": ["log_tab"]},
    }

    if config is not None:
        log_file = config.log_file or ""
        log_file_is_stdout = log_file in {"-", "/dev/stdout"}
        log_level = config.log_level.upper()

        # Configure file handler
        if log_file and not log_file_is_stdout:
            log_config["handlers"]["file"] = {
                "level": log_level,
                "class": "logging.FileHandler",
                "filename": Path(config.log_file).expanduser(),
                "formatter": "file_format",
            }
            log_config["loggers"]["euporie"]["handlers"].append("file")

        # Configure stdout handler
        if log_file_is_stdout:
            stdout_level = log_level
        else:
            stdout_level = config.log_level_stdout.upper()
        log_config["handlers"]["stdout"]["level"] = stdout_level
        if syntax_theme := config.syntax_theme:
            log_config["handlers"]["stdout"]["pygments_theme"] = syntax_theme

        # Configure euporie logger
        log_config["loggers"]["euporie"]["level"] = log_level

        # Update log_config based on additional config dict provided
        if config.log_config:
            dict_merge(log_config, config.log_config)

    # Configure the logger
    # Pytype used TypedDicts to validate the dictionary structure, but I cannot get
    # this to work for some reason...
    logging.config.dictConfig(log_config)  # type: ignore

    # Capture warnings so they show up in the logs
    logging.captureWarnings(True)

    # Log uncaught exceptions
    sys.excepthook = handle_exception



"""Tab for use in euporie notebook editor."""

from euporie.core.convert.mime import MIME_FORMATS
from euporie.core.tabs import _TAB_REGISTRY, TabRegistryEntry

_TAB_REGISTRY.extend(
    [
        TabRegistryEntry(
            path="euporie.notebook.tabs.display:DisplayTab",
            name="File Viewer",
            mime_types=set(MIME_FORMATS.keys()),
        ),
        TabRegistryEntry(
            path="euporie.notebook.tabs.edit:EditorTab",
            name="Text Editor",
            mime_types={"text/*"},
            weight=1,
        ),
        TabRegistryEntry(
            path="euporie.notebook.tabs.json:JsonTab",
            name="JSON Viewer",
            mime_types={"*json"},
            file_extensions={".json": None},
        ),
        TabRegistryEntry(
            path="euporie.notebook.tabs.notebook:Notebook",
            name="Notebook Editor",
            mime_types={"application/x-ipynb+json"},
            file_extensions={".ipynb": None},
            weight=3,
        ),
        TabRegistryEntry(
            path="euporie.web.tabs.web:WebTab",
            name="Web Viewer",
            mime_types={"text/html", "text/markdown"},
            weight=2,
        ),
    ]
)

try:
    from jupytext.formats import NOTEBOOK_EXTENSIONS
except ModuleNotFoundError:
    pass
else:
    for entry in _TAB_REGISTRY:
        if entry.path == "euporie.notebook.tabs.notebook:Notebook":
            entry.file_extensions.update(dict.fromkeys(NOTEBOOK_EXTENSIONS))
            break



"""Tabs for use in euporie notebook editor."""

from euporie.core.tabs import _TAB_REGISTRY, TabRegistryEntry

_TAB_REGISTRY.extend(
    [
        TabRegistryEntry(
            path="euporie.web.tabs.web:WebTab",
            name="Web Viewer",
            mime_types={"text/html", "text/markdown"},
            weight=2,
        ),
    ]
)



"""Define the global command toolbar."""

from __future__ import annotations

import logging
import re
from functools import lru_cache
from typing import TYPE_CHECKING

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion.base import Completer, Completion
from prompt_toolkit.filters import buffer_has_focus, has_focus, vi_navigation_mode
from prompt_toolkit.key_binding.vi_state import InputMode
from prompt_toolkit.layout.containers import ConditionalContainer, Container, Window
from prompt_toolkit.layout.controls import (
    BufferControl,
)
from prompt_toolkit.layout.processors import BeforeInput, HighlightSelectionProcessor
from prompt_toolkit.lexers import SimpleLexer
from prompt_toolkit.validation import Validator

from euporie.core.app.current import get_app
from euporie.core.bars import COMMAND_BAR_BUFFER
from euporie.core.commands import add_cmd, commands, get_cmd
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Unpack

    from prompt_toolkit.completion.base import CompleteEvent
    from prompt_toolkit.document import Document
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent

    from euporie.core.commands import Command

log = logging.getLogger(__name__)


@lru_cache
def _parse_cmd(text: str) -> tuple[Command | None, str]:
    """Parse a command line to command and arguments.

    Command names cannot start with digits, so lines staring with digits have an empty
    command string (this is used for the go-to-cell/go-to-line shortcuts).
    """
    if match := re.fullmatch(r"^(?P<cmd>[^\d][^\s]*|)\s*(?P<args>.*)$", text):
        cmd, args = match.groups()
    else:
        cmd, args = "", ""
    try:
        return get_cmd(cmd), args
    except KeyError:
        return None, args


class CommandCompleter(Completer):
    """Completer of commands."""

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        """Complete registered commands."""
        prefix = document.text
        found_so_far: set[Command] = set()
        for alias, command in commands.items():
            if (
                alias.startswith(prefix)
                and command not in found_so_far
                and not command.hidden()
            ):
                yield Completion(
                    command.name,
                    start_position=-len(prefix),
                    display=command.name,
                    display_meta=command.description,
                )
                found_so_far.add(command)


class CommandBar:
    """Command mode toolbar.

    A modal editor like toolbar to allow entry of commands.
    """

    def __init__(self) -> None:
        """Create a new command bar instance."""
        self.buffer = Buffer(
            completer=CommandCompleter(),
            complete_while_typing=True,
            name=COMMAND_BAR_BUFFER,
            multiline=False,
            accept_handler=self._accept,
            validator=Validator.from_callable(
                validate_func=self._validate,
                error_message="Command not recognised",
                move_cursor_to_end=True,
            ),
        )
        self.control = BufferControl(
            buffer=self.buffer,
            lexer=SimpleLexer(style="class:toolbar.text"),
            input_processors=[
                BeforeInput(":", style="class:toolbar.title"),
                HighlightSelectionProcessor(),
            ],
            include_default_input_processors=False,
            key_bindings=load_registered_bindings(
                "euporie.core.bars.command:CommandBar",
                config=get_app().config,
            ),
        )
        self.window = Window(
            self.control,
            height=1,
            style="class:toolbar",
        )

        self.container = ConditionalContainer(
            content=self.window,
            filter=has_focus(self.buffer),
        )

    def _validate(self, text: str) -> bool:
        """Verify that a valid command has been entered."""
        cmd, _args = _parse_cmd(text)
        return bool(cmd)

    def _accept(self, buffer: Buffer) -> bool:
        """Return value determines if the text is kept."""
        app = get_app()
        app.vi_state.input_mode = InputMode.NAVIGATION
        app.layout.focus_last()
        text = buffer.text.strip()
        cmd, args = _parse_cmd(text)
        if cmd:
            cmd.run(args)
        return False

    def __pt_container__(self) -> Container:
        """Magic method for widget container."""
        return self.container

    register_bindings(
        {
            "euporie.core.app.app:BaseApp": {
                "activate-command-bar": ":",
                "activate-command-bar-alt": "A-:",
                "activate-command-bar-shell": "!",
                "activate-command-bar-shell-alt": "A-!",
            },
            "euporie.core.bars.command:CommandBar": {
                "deactivate-command-bar": ["escape", "c-c"],
            },
        }
    )

    @staticmethod
    @add_cmd(name="activate-command-bar-alt", hidden=True)
    @add_cmd(filter=~buffer_has_focus | vi_navigation_mode)
    def _activate_command_bar(event: KeyPressEvent) -> None:
        """Enter command mode."""
        event.app.layout.focus(COMMAND_BAR_BUFFER)
        event.app.vi_state.input_mode = InputMode.INSERT

    @staticmethod
    @add_cmd(filter=~buffer_has_focus)
    @add_cmd(name="activate-command-bar-shell-alt", hidden=True)
    def _activate_command_bar_shell(event: KeyPressEvent) -> None:
        """Enter command mode."""
        app = event.app
        layout = app.layout
        layout.focus(COMMAND_BAR_BUFFER)
        app.vi_state.input_mode = InputMode.INSERT
        if isinstance(control := layout.current_control, BufferControl):
            buffer = control.buffer
            buffer.text = "shell "
            buffer.cursor_position = 6

    @staticmethod
    @add_cmd(hidden=True)
    def _deactivate_command_bar(event: KeyPressEvent) -> None:
        """Exit command mode."""
        app = event.app
        layout = app.layout
        layout.focus(COMMAND_BAR_BUFFER)
        if isinstance(control := layout.current_control, BufferControl):
            app.vi_state.input_mode = InputMode.NAVIGATION
            buffer = control.buffer
            buffer.reset()
            app.layout.focus_previous()

    @staticmethod
    @add_cmd(aliases=["shell"])
    async def _run_shell_command(
        event: KeyPressEvent, *cmd_arg: Unpack[tuple[str]]
    ) -> None:
        """Run system command."""
        command = " ".join(str(x) for x in cmd_arg)
        if command:
            await event.app.run_system_command(
                command,
                display_before_text=[("bold", "$ "), ("", f"{command}\n")],
            )



"""Defines a simple LSP client."""

from __future__ import annotations

import asyncio
import json
import logging
import os
import threading
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.utils import Event

if TYPE_CHECKING:
    from collections.abc import Coroutine, Sequence
    from concurrent.futures import Future
    from typing import Any, Callable


log = logging.getLogger(__name__)


def _setup_loop(loop: asyncio.AbstractEventLoop) -> None:
    """Run a new event loop (intended to start a loop running in a thread."""
    asyncio.set_event_loop(loop)
    loop.run_forever()


_COMPLETION_TYPES = {
    1: "text",
    2: "method",
    3: "function",
    4: "constructor",
    5: "field",
    6: "variable",
    7: "class",
    8: "interface",
    9: "module",
    10: "property",
    11: "unit",
    12: "value",
    13: "enum",
    14: "keyword",
    15: "snippet",
    16: "color",
    17: "file",
    18: "reference",
    19: "folder",
    20: "enum_member",
    21: "constant",
    22: "struct",
    23: "event",
    24: "operator",
    25: "type_parameter",
}


def range_to_slice(
    start_line: int, start_char: int, end_line: int, end_char: int, text: str
) -> slice:
    """Convert a line/character ranger to a slice."""
    start_pos = start_char
    end_pos = end_char
    lines = text.splitlines(keepends=True)
    for i, line in enumerate(lines):
        len_line = len(line)
        if i < start_line:
            start_pos += len_line
        if i < end_line:
            end_pos += len_line
        else:
            break
    return slice(start_pos, end_pos)


class LspCell(NamedTuple):
    """An LSP client's representation of a cell."""

    id: str
    idx: int
    path: Path
    kind: str
    language: str
    text: str
    execution_count: int
    metadata: dict[str, Any] | None = None


class LspClient:
    """A client for communicating with LSP servers."""

    # An event loop for running the LSP clients
    loop = asyncio.new_event_loop()
    # A thread used to run the LSP client event loop
    thread = threading.Thread(target=_setup_loop, args=(loop,))
    thread.daemon = True

    def __init__(
        self,
        name: str,
        command: str,
        languages: Sequence[str] | None = None,
        settings: dict | None = None,
    ) -> None:
        """Initialize the LSP client."""
        self.name = name
        self.command = command
        self.languages = set(languages) if languages else set()
        self.settings = settings or {}
        self.started = False

        # The current RPC message ID
        self._msg_id = 0
        # References to callbacks for expected message responses
        self._msg_callbacks: dict[int, Callable[[Any], Coroutine[Any, Any, None]]] = {}
        # Change version numbers for opened documents
        self._doc_versions: dict[Path, int] = {}
        # Futures for change events, allowing debouncing changes
        self._change_futures: dict[Path, Future[None]] = {}
        # Holds diagnostic reports (NOTE: uses URI string from response as key)
        self.reports: dict[str, list] = {}

        self.initialized = asyncio.Event()

        # Server capability flags. None = not sure yet
        self.can_hover: bool | None = None
        self.can_format: bool | None = None
        self.can_complete: bool | None = None
        self.can_signature: bool | None = None
        # `open_doc`, `change_doc` and `close_doc` are mandatory for LSP servers
        self.can_open_doc: bool | None = True
        self.can_close_doc: bool | None = True
        self.can_change_doc: bool | None = True
        self.can_change_doc_incr: bool | None = None
        self.can_save_doc: bool | None = None
        self.can_will_save_doc: bool | None = None
        self.can_open_nb: bool | None = None
        self.can_change_nb: bool | None = None
        self.can_change_nb_langs: list[str] = []

        self._request_handlers = {
            "window/logMessage": self.log_message,
            "textDocument/publishDiagnostics": self.diagnostics,
        }

        self.on_diagnostics = Event(self)

        # Ensure LSP servers get shutdown at exit
        self.on_exit = Event(self)

    def __repr__(self) -> str:
        """Return a string representation of the LSP client."""
        return f"{self.__class__.__name__}(name={self.name})"

    @property
    def next_msg_id(self) -> int:
        """Get and increment the RPC message ID."""
        self._msg_id += 1
        return self._msg_id

    def start(self, root: Path | None = None) -> None:
        """Start the LSP server."""
        if not self.started:
            try:
                self.thread.start()
            except RuntimeError:
                pass
            # Start lsp server process and monitor it on event loop in thread
            asyncio.run_coroutine_threadsafe(self.start_(root), self.loop).result(
                timeout=30
            )

    async def start_(self, root: Path | None = None) -> None:
        """Launch the LSP server subprocess."""
        self.process = await asyncio.create_subprocess_exec(
            *self.command,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            # stderr=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
        )
        self.monitor_stdout_task = self.loop.create_task(self._monitor_stdout())
        self.monitor_stderr_task = self.loop.create_task(self._monitor_stderr())
        self.started = True

        # Tell the server to initialize
        await self.initialize(root)

    async def _monitor_stderr(self) -> None:
        """Log LSP server standard error output."""
        while stderr := self.process.stderr:
            line = await stderr.readline()
            if line:
                log.info(line.decode().rstrip())
            else:
                break

    async def _monitor_stdout(self) -> None:
        """Retrieve and parse responses."""
        buffer = b""
        length = 0
        # mime: str = None
        charset: str = "utf-8"

        while stdout := self.process.stdout:
            if buffer.startswith(b"Content-Length: "):
                line_end = buffer.find(b"\r\n")
                _, _, _length = buffer[:line_end].partition(b": ")
                length = int(_length)
                buffer = buffer[line_end + 2 :]
                continue

            if buffer.startswith(b"Content-Type: "):
                line_end = buffer.find(b"\r\n")
                _, _, _content_type = buffer[:line_end].partition(b": ")
                # TODO - get mime & encoding
                # mime =
                # charset =
                buffer = buffer[line_end + 2 :]
                continue

            if buffer.startswith(b"\r\n"):
                buffer = buffer[2:]
                continue

            if length and len(buffer) >= length:
                content = buffer[:length]
                buffer = buffer[length:]
                length = 0
                await self.process_msg(json.loads(content.decode(charset)))
                continue

            else:
                # Read a (large) chunk of the server's stdout
                line = await stdout.read(1024 * 1024)
                if line:
                    buffer += line
                else:
                    break

    async def process_msg(self, data: dict) -> None:
        """Process an incoming message from the LSP server."""
        # log.debug("%s received message %s", self, data)
        if msg_id := data.get("id"):
            if callback := self._msg_callbacks.get(msg_id):
                try:
                    await callback(data.get("result", {}))
                except Exception:
                    log.exception("An error occurred executing the LSP callback")
                finally:
                    del self._msg_callbacks[msg_id]

                return

        elif method := data.get("method"):
            # A request from the server
            if handler := self._request_handlers.get(method):
                handler(data.get("params", {}))

            else:
                log.info(data)

        else:
            # Some other response from the server
            log.info(data)

    async def send_msg(
        self,
        method: str,
        params: dict[str, Any] | None = None,
        cb: Callable[[Any], Coroutine[Any, Any, None]] | None = None,
    ) -> int:
        """Send a message to the LSP server, returning the sent message's ID."""
        msg_id = self.next_msg_id
        msg: dict[str, Any] = {
            "jsonrpc": "2.0",
            "method": method,
        }
        if params is not None:
            msg["params"] = params
        if cb:
            msg["id"] = msg_id
        body = json.dumps(msg, separators=(",", ":"))
        # log.debug("Sending %s message %s", self, body)
        data = f"Content-Length: {len(body)}\r\n\r\n{body}"
        if stdin := self.process.stdin:
            stdin.write(data.encode())
            await stdin.drain()
        else:
            log.error("Unexpectedly lost connection with %s", self)
        if callable(cb):
            self._msg_callbacks[msg_id] = cb
        return msg_id

    async def initialize(self, root: Path | None = None) -> None:
        """Initiate the LSP server."""
        if not self.initialized.is_set():
            if root is None:
                root = Path.cwd()
            await self.send_msg(
                method="initialize",
                params={
                    "trace": "off",
                    "processId": os.getpid(),
                    "capabilities": {
                        "workspace": {
                            "applyEdit": False,
                            "fileOperations": {
                                "dynamicRegistration": False,
                                "didCreate": False,
                                "willCreate": False,
                                "didRename": False,
                                "willRename": False,
                                "didDelete": False,
                                "willDelete": False,
                            },
                        },
                        "textDocument": {
                            "willSave": True,
                            "didSave": True,
                        },
                        "notebookDocument": {
                            "synchronization": {"executionSummarySupport": True}
                        },
                        "documentSelector": [
                            {"scheme": "file"},
                            {"scheme": "untitled"},
                        ],
                    },
                    "rootUri": root.as_uri(),
                    "clientInfo": {"name": "euporie", "version": "2.7.0"},
                },
                cb=self._on_initialize_response,
            )

    async def _on_initialize_response(self, rsp: dict[str, Any]) -> None:
        server_info = rsp.get("serverInfo", {})
        self.server_name = server_info.get("name")
        self.server_version = server_info.get("version")

        capabilities = rsp.get("capabilities", {})

        _change_options = capabilities.get("textDocumentSync", {})
        if isinstance(_change_options, int):
            _change_kind = _change_options
            _change_options = {}
        else:
            _change_kind = _change_options.get("change")
        if isinstance(_change_options.get("save"), bool):
            _save_options = {}
        else:
            _save_options = _change_options.get("save", {})

        self.can_hover = (
            "hoverProvider" in capabilities
            and capabilities["hoverProvider"] is not False
        )
        self.can_format = (
            "documentFormattingProvider" in capabilities
            and capabilities["documentFormattingProvider"] is not False
        )
        self.can_complete = (
            "completionProvider" in capabilities
            and capabilities["completionProvider"] is not False
        )
        self.can_signature = (
            "signatureHelpProvider" in capabilities
            and capabilities["signatureHelpProvider"] is not False
        )

        self.can_change_doc = "textDocumentSync" in capabilities and (
            _change_kind is None or _change_kind > 0
        )
        self.can_change_doc_incr = _change_kind == 2
        self.can_save_doc = "save" in _change_options
        self.can_save_doc_text = _save_options.get("includeText", True)
        self.can_will_save_doc = "willSave" in _change_options

        _nb_change_options = capabilities.get("notebookDocumentSync", {})
        self.can_open_nb = self.can_change_nb = self.can_close_nb = bool(
            _nb_change_options
        )
        # self.can_change_nb = any(
        #     "notebook" in selector
        #     for selector in _nb_change_options.get("notebookSelector", [])
        # )
        self.can_change_nb_langs = [
            cell_selector.get("language")
            for nb_selector in _nb_change_options.get("notebookSelector", [])
            for cell_selector in nb_selector.get("cells", [])
            if cell_selector.get("language")
        ]
        self.can_save_nb = _nb_change_options.get("save")

        # for name, value in rsp.get("capabilities", {}).items():
        #     log.debug("%s: %s: %s", self, name, value)

        await self.send_msg(method="initialized", params={})
        self.initialized.set()

        # Send settings
        await self.send_msg(
            method="workspace/didChangeConfiguration",
            params={"settings": self.settings},
        )

    def exit(self) -> Future:
        """Tell the server to exit."""
        return asyncio.run_coroutine_threadsafe(self.exit_(), self.loop)

    async def exit_(self, timeout: int = 1) -> None:
        """Tell the server to exit."""
        log.debug("%s exiting", self)
        self.on_exit.fire()
        event = asyncio.Event()

        async def _cb(result: dict | None) -> None:
            await self.send_msg(method="exit")
            event.set()

        await self.send_msg(method="shutdown", cb=_cb)

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for %s shutdown response", self)

        try:
            await asyncio.wait_for(self.process.wait(), timeout=1)
        except TimeoutError:
            self.process.kill()
        log.debug("%s exited", self)

    def open_doc(self, path: Path, language: str, text: str) -> None:
        """Tell the server we have opened a document file."""
        # "An open notification must not be sent more than once without a corresponding
        #  close notification send before"
        if path in self._doc_versions:
            return
        # else:
        #     log.debug("%s opening %s", self, path)
        if self.languages and language not in self.languages:
            return
        self._doc_versions[path] = 0
        asyncio.run_coroutine_threadsafe(
            self.send_msg(
                method="textDocument/didOpen",
                params={
                    "textDocument": {
                        "uri": path.as_uri(),
                        "languageId": language.lower(),
                        "version": 0,
                        "text": text,
                    }
                },
            ),
            self.loop,
        )

    def open_nb(
        self, path: Path, cells: Sequence[LspCell], metadata: dict[str, Any] | None
    ) -> None:
        """Tell the server we have opened a notebook file."""
        if not self.can_open_nb:
            return
        self._doc_versions[path] = 0
        params: dict[str, Any] = {
            "notebookDocument": {
                "uri": path.as_uri(),
                "notebookType": "jupyter-notebook",
                "version": 0,
                # "metadata": metadata or {},
                "cells": [],
            },
            "cellTextDocuments": [],
        }
        langs = self.can_change_nb_langs
        lsp_notebook_cell: dict[str, Any]
        for cell in cells:
            if langs and cell.language not in langs:
                continue
            self._doc_versions[cell.path] = 0
            uri = cell.path.as_uri()
            params["notebookDocument"]["cells"].append(
                lsp_notebook_cell := {
                    "kind": 2 if cell.kind == "code" else 1,
                    "document": uri,
                    # "metadata": cell.metadata,
                }
            )
            if cell.execution_count:
                lsp_notebook_cell["executionSummary"] = {
                    "executionOrder": cell.execution_count,
                    # "success": True,
                }
            params["cellTextDocuments"].append(
                {
                    "uri": uri,
                    "languageId": cell.language.lower(),
                    "version": 0,
                    "text": cell.text,
                }
            )
        asyncio.run_coroutine_threadsafe(
            self.send_msg(
                method="notebookDocument/didOpen",
                params=params,
            ),
            self.loop,
        )

    def change_doc(
        self, path: Path, language: str, text: str, debounce: bool = True
    ) -> None:
        """Tell the server we have changed a document file."""
        if (langs := self.languages) and language not in langs:
            return

        if path in self._change_futures:
            self._change_futures[path].cancel()
            del self._change_futures[path]

        async def _notify_change() -> None:
            if debounce:
                await asyncio.sleep(0.25)
            self._doc_versions[path] += 1
            await self.send_msg(
                method="textDocument/didChange",
                params={
                    "textDocument": {
                        "uri": path.as_uri(),
                        "version": self._doc_versions[path],
                    },
                    "contentChanges": [
                        {
                            # "range": {
                            #     "start": {"line": 0, "character": 0},
                            #     "end": {"line": 6, "character": 0},
                            # },
                            "text": text,
                        }
                    ],
                },
            )

        self._change_futures[path] = asyncio.run_coroutine_threadsafe(
            _notify_change(), self.loop
        )
        if not debounce:
            self._change_futures[path].result()

    def change_nb_meta(self, path: Path, metadata: dict[str, Any]) -> None:
        """Notify server of a change in a notebook metadata."""
        if not self.can_change_nb:
            return
        # Update versions
        self._doc_versions[path] += 1
        asyncio.run_coroutine_threadsafe(
            self.send_msg(
                method="notebookDocument/didChange",
                params={
                    "notebookDocument": {
                        "uri": path.as_uri(),
                        "version": self._doc_versions[path],
                    },
                    "change": {"metadata": metadata},
                },
            ),
            self.loop,
        )

    def change_nb_add(self, path: Path, cells: Sequence[LspCell]) -> None:
        """Notify server of cells added to notebook."""
        if not self.can_change_nb:
            return
        # Update notebook version
        self._doc_versions[path] += 1
        # Set RPC parameters
        change: dict[str, Any]  # TODO - use typed dicts?
        params = {
            "notebookDocument": {
                "uri": path.as_uri(),
                "version": self._doc_versions[path],
            },
            "change": {
                "cells": (
                    change := {
                        "structure": {
                            "array": {
                                "start": min(cell.idx for cell in cells),
                                "deleteCount": 0,
                                "cells": [],
                            },
                            "didOpen": [],
                        },
                        "data": [],
                    }
                )
            },
        }
        langs = self.can_change_nb_langs
        lsp_notebook_cell: dict[str, Any]
        for cell in cells:
            # If specified, only sync cells with the languages the server has requested
            if langs and cell.language not in langs:
                continue
            self._doc_versions[cell.path] = 0
            lsp_notebook_cell = {
                "kind": 2 if cell.kind == "code" else 1,
                "document": cell.path.as_uri(),
                # "metadata": cell.metadata,
            }
            if cell.execution_count:
                lsp_notebook_cell["executionSummary"] = {
                    "executionOrder": cell.execution_count,
                    # "success": True,
                }
            change["structure"]["array"]["cells"].append(lsp_notebook_cell)
            change["structure"]["didOpen"].append(
                {
                    "uri": cell.path.as_uri(),
                    "languageId": cell.language,
                    "version": self._doc_versions[cell.path],
                    "text": cell.text,
                }
            )
            change["data"].append(lsp_notebook_cell)
        # Run RPC call
        asyncio.run_coroutine_threadsafe(
            self.send_msg(
                method="notebookDocument/didChange",
                params=params,
            ),
            self.loop,
        )

    def change_nb_edit(
        self, path: Path, cells: Sequence[LspCell], debounce: bool = True
    ) -> None:
        """Notify server of a change in  notebook cells."""
        if not self.can_change_nb:
            return

        if path in self._change_futures:
            self._change_futures[path].cancel()
            del self._change_futures[path]

        async def _notify_change() -> None:
            if debounce:
                await asyncio.sleep(0.25)
            self._doc_versions[path] += 1

            change: dict[str, list[dict]]
            params = {
                "notebookDocument": {
                    "uri": path.as_uri(),
                    "version": self._doc_versions[path],
                },
                "change": {"cells": (change := {"data": [], "textContent": []})},
            }
            langs = self.can_change_nb_langs
            lsp_notebook_cell: dict[str, Any]
            for cell in cells:
                self._doc_versions[cell.path] += 1
                # If specified, only sync cells with the languages the server has requested
                if langs and cell.language not in langs:
                    continue
                change["data"].append(
                    lsp_notebook_cell := {
                        "kind": 2 if cell.kind == "code" else 1,
                        "document": cell.path.as_uri(),
                        # "metadata": cell.metadata,
                    }
                )
                if cell.execution_count:
                    lsp_notebook_cell["executionSummary"] = {
                        "executionOrder": cell.execution_count,
                        # "success": True,
                    }
                change["textContent"].append(
                    {
                        "document": {
                            "uri": cell.path.as_uri(),
                            "version": self._doc_versions[cell.path],
                        },
                        "changes": [{"text": cell.text}],
                    }
                )

            # Run RPC call
            await self.send_msg(
                method="notebookDocument/didChange",
                params=params,
            )

        self._change_futures[path] = asyncio.run_coroutine_threadsafe(
            _notify_change(), self.loop
        )
        if not debounce:
            self._change_futures[path].result()

    def change_nb_delete(self, path: Path, cells: Sequence[LspCell]) -> None:
        """Notify server of cells removed from a notebook."""
        if not self.can_change_nb:
            return
        # Update versions
        self._doc_versions[path] += 1
        langs = self.can_change_nb_langs
        for cell in cells:
            # If specified, only sync cells with the languages the server has requested
            if langs and cell.language not in langs:
                continue
            self._cleanup_path(cell.path)
        # Set RPC parameters
        asyncio.run_coroutine_threadsafe(
            self.send_msg(
                method="notebookDocument/didChange",
                params={
                    "notebookDocument": {
                        "uri": path.as_uri(),
                        "version": self._doc_versions[path],
                    },
                    "change": {
                        "cells": {
                            "structure": {
                                "array": {
                                    "start": min(cell.idx for cell in cells),
                                    "deleteCount": len(cells),
                                },
                                "didClose": [
                                    {"uri": cell.path.as_uri()} for cell in cells
                                ],
                            }
                        }
                    },
                },
            ),
            self.loop,
        )

    def will_save_doc(self, path: Path) -> None:
        """Tell the server we will save a document file."""
        if self.can_will_save_doc:
            asyncio.run_coroutine_threadsafe(
                self.send_msg(
                    method="textDocument/willSave",
                    params={
                        "textDocument": {"uri": path.as_uri()},
                        "reason": 1,  # Manual save
                    },
                ),
                self.loop,
            )

    def will_save_nb(self, path: Path) -> None:
        """Tell the server we will save a notebook file."""

    def save_doc(self, path: Path, text: str) -> None:
        """Tell the server we saved a text document."""
        if self.can_save_doc:
            asyncio.run_coroutine_threadsafe(
                self.send_msg(
                    method="textDocument/didSave",
                    params={
                        "textDocument": {"uri": path.as_uri()},
                        **({"text": text} if self.can_save_doc_text else {}),
                    },
                ),
                self.loop,
            )

    def save_nb(self, path: Path) -> None:
        """Tell the server we saved a notebook."""
        asyncio.run_coroutine_threadsafe(
            self.send_msg(
                method="notebookDocument/didSave",
                params={
                    "notebookDocument": {"uri": path.as_uri()},
                },
            ),
            self.loop,
        )

    def _cleanup_path(self, path: Path) -> None:
        try:
            del self._doc_versions[path]
        except KeyError:
            pass
        try:
            del self._change_futures[path]
        except KeyError:
            pass
        try:
            del self.reports[path.as_uri()]
        except KeyError:
            pass

    def close_doc(self, path: Path) -> None:
        """Tell the server we have closed a document file."""
        # log.debug("%s closing %s", self, path)
        self._cleanup_path(path)
        asyncio.run_coroutine_threadsafe(
            self.send_msg(
                method="textDocument/didClose",
                params={
                    "textDocument": {"uri": path.as_uri()},
                },
            ),
            self.loop,
        )

    def close_nb(self, path: Path, cells: Sequence[LspCell]) -> None:
        """Tell the server we closed a notebook document."""
        # log.debug("%s closing %s", self, path)
        params = {"notebookDocument": {"uri": path.as_uri()}, "cellTextDocuments": []}
        for cell in cells:
            self._cleanup_path(cell.path)
            # params["cellTextDocuments"].append({"uri": cell_path.as_uri()})
        self._cleanup_path(path)
        asyncio.run_coroutine_threadsafe(
            self.send_msg(method="notebookDocument/didClose", params=params), self.loop
        )

    def complete(
        self,
        path: Path,
        line: int,
        char: int,
        timeout: int = 5,
    ) -> list[dict[str, str]]:
        """Trigger a LSP completion request."""
        if not self.can_complete:
            return []
        return asyncio.run_coroutine_threadsafe(
            self.complete_(path, line, char, timeout), self.loop
        ).result()

    async def complete_(
        self, path: Path, line: int, char: int, timeout: int = 5
    ) -> list[dict[str, Any]]:
        """Request a completion from the LSP server."""
        if not self.can_complete:
            return []
        results = []
        event = asyncio.Event()

        async def _cb(rsp: dict[str, Any]) -> None:
            for item in rsp.get("items", []):
                results.append(
                    {
                        "text": item["insertText"],
                        "display": item.get("label"),
                        # "start_position": rel_start_position,
                        "display_meta": _COMPLETION_TYPES.get(item.get("kind")),
                    }
                )
            event.set()

        await self.send_msg(
            method="textDocument/completion",
            params={
                "textDocument": {"uri": path.as_uri()},
                "position": {"line": line, "character": char},
                "context": {
                    "triggerKind": 1  # Manually triggered
                },
            },
            cb=_cb,
        )

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for %s completion response", self)

        return results

    def hover(self, path: Path, line: int, char: int, timeout: int = 1) -> dict:
        """Trigger a LSP hover request."""
        if not self.can_hover:
            return {}
        return asyncio.run_coroutine_threadsafe(
            self.hover_(path, line, char, timeout), self.loop
        ).result()

    async def hover_(self, path: Path, line: int, char: int, timeout: int = 5) -> dict:
        """Request hover text from the LSP server."""
        if not self.can_hover:
            return {}
        result = {}
        event = asyncio.Event()

        async def _cb(rsp: dict[str, Any]) -> None:
            if rsp:
                contents = rsp.get("contents", {})
                mime = "text/markdown"
                if isinstance(contents, str):
                    data = contents
                elif "language" in contents:
                    language = contents.get("language", "")
                    value = contents.get("value", "")
                    data = f"```${language}\n${value}\n```"
                elif "kind" in contents:
                    if contents["kind"] == "plaintext":
                        mime = "text/plain"
                    data = contents.get("value", "")
                if data:
                    result.update({mime: data})
            event.set()

        await self.send_msg(
            method="textDocument/hover",
            params={
                "textDocument": {"uri": path.as_uri()},
                "position": {"line": line, "character": char},
            },
            cb=_cb,
        )
        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for %s hover response", self)

        return result

    def signature(
        self, path: Path, line: int, char: int, timeout: int = 1
    ) -> dict[str, object] | None:
        """Trigger a LSP signature request."""
        if not self.can_signature:
            return None
        return asyncio.run_coroutine_threadsafe(
            self.signature_(path, line, char, timeout=timeout), self.loop
        ).result()

    async def signature_(
        self, path: Path, line: int, char: int, timeout: int = 5
    ) -> dict[str, object] | None:
        """Request a signature from the LSP server."""
        if not self.can_signature:
            return None
        result = {}
        event = asyncio.Event()

        async def _cb(rsp: dict[str, Any]) -> None:
            mime = "text/markdown"
            contents = rsp.get("contents", {})
            if isinstance(contents, str):
                data = contents
            elif "language" in contents:
                language = contents.get("language", "")
                value = contents.get("value", "")
                data = f"```${language}\n${value}\n```"
            elif "kind" in contents:
                if contents["kind"] == "plaintext":
                    mime = "text/plain"
                data = contents.get("value", "")
            result.update({"found": True, "data": {mime: data}})
            event.set()

        await self.send_msg(
            method="textDocument/signatureHelp",
            params={
                "textDocument": {
                    "uri": path.as_uri(),
                    "version": self._doc_versions[path],
                },
                "position": {"line": line, "character": char},
            },
            cb=_cb,
        )
        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for %s signature response", self)

        return result

    def format(
        self, path: Path, tab_size: int = 4, spaces: bool = True, timeout: int = 1
    ) -> list[dict] | None:
        """Request LSP document formatting."""
        if not self.can_format:
            return []
        return asyncio.run_coroutine_threadsafe(
            self.format_(path, timeout=timeout), self.loop
        ).result()

    async def format_(
        self, path: Path, tab_size: int = 4, spaces: bool = True, timeout: int = 5
    ) -> list[dict] | None:
        """Trigger a formatting request from the LSP."""
        if not self.can_format:
            return []
        result = []
        event = asyncio.Event()

        async def _cb(rsp: list) -> None:
            if rsp:
                result.extend(rsp)
            event.set()

        await self.send_msg(
            method="textDocument/formatting",
            params={
                "textDocument": {
                    "uri": path.as_uri(),
                    "version": self._doc_versions[path],
                },
                "options": {"tabSize": tab_size, "insertSpaces": spaces},
            },
            cb=_cb,
        )
        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for %s format response", self)

        return result

    def log_message(self, params: dict[str, Any]) -> None:
        """Send a log message from the server to the log."""
        if msg := params.get("message"):
            msg_type = params.get("type", 2)
            level = (5 - msg_type) * 10
            log.log(level, msg)

    def diagnostics(self, params: dict[str, Any]) -> None:
        """Receive a diagnstics report from the server."""
        uri: str
        if uri := params.get("uri", ""):
            self.reports[uri] = params.get("diagnostics", [])
            self.on_diagnostics.fire()


KNOWN_LSP_SERVERS: dict[str, dict[str, Any]] = {
    # Custom additions
    "typos": {"command": ["typos-lsp"], "languages": set()},
    # Helix's known LSPs
    "als": {"command": ["als"]},
    "ada-language-server": {"command": ["ada_language_server"], "languages": {"ada"}},
    "ada-gpr-language-server": {
        "command": ["ada_language_server", "--language-gpr"],
        "languages": {"ada"},
    },
    "angular": {
        "command": [
            "ngserver",
            "--stdio",
            "--tsProbeLocations",
            ".",
            "--ngProbeLocations",
            ".",
        ]
    },
    "awk-language-server": {"command": ["awk-language-server"], "languages": {"awk"}},
    "bash-language-server": {
        "command": ["bash-language-server", "start"],
        "languages": {"bash"},
    },
    "bass": {"command": ["bass", "--lsp"], "languages": {"bass"}},
    "bicep-langserver": {"command": ["bicep-langserver"], "languages": {"bicep"}},
    "bitbake-language-server": {
        "command": ["bitbake-language-server"],
        "languages": {"bitbake"},
    },
    "bufls": {"command": ["bufls", "serve"], "languages": {"protobuf"}},
    "cairo-language-server": {
        "command": ["cairo-language-server"],
        "languages": {"cairo"},
    },
    "circom-lsp": {"command": ["circom-lsp"], "languages": {"circom"}},
    "cl-lsp": {"command": ["cl-lsp", "stdio"], "languages": {"common-lisp"}},
    "clangd": {"command": ["clangd"], "languages": {"opencl", "cpp", "c"}},
    "clojure-lsp": {"command": ["clojure-lsp"], "languages": {"clojure"}},
    "cmake-language-server": {
        "command": ["cmake-language-server"],
        "languages": {"cmake"},
    },
    "crystalline": {"command": ["crystalline", "--stdio"], "languages": {"crystal"}},
    "cs": {
        "command": ["cs", "launch", "--contrib", "smithy-language-server", "--", "0"],
        "languages": {"smithy"},
    },
    "csharp-ls": {"command": ["csharp-ls"]},
    "cuelsp": {"command": ["cuelsp"], "languages": {"cue"}},
    "dart": {
        "command": ["dart", "language-server", "--client-id=helix"],
        "languages": {"dart"},
    },
    "dhall-lsp-server": {"command": ["dhall-lsp-server"], "languages": {"dhall"}},
    "docker-langserver": {
        "command": ["docker-langserver", "--stdio"],
        "languages": {"dockerfile"},
    },
    "docker-compose-langserver": {
        "command": ["docker-compose-langserver", "--stdio"],
        "languages": {"docker-compose"},
    },
    "dot-language-server": {
        "command": ["dot-language-server", "--stdio"],
        "languages": {"dot"},
    },
    "earthlyls": {"command": ["earthlyls"], "languages": {"earthfile"}},
    "elixir-ls": {
        "config": {"elixirLS": {"dialyzerEnabled": False}},
        "command": ["elixir-ls"],
        "languages": {"heex", "elixir"},
    },
    "elm-language-server": {"command": ["elm-language-server"], "languages": {"elm"}},
    "elp": {"command": ["elp", "server"], "languages": {"erlang"}},
    "elvish": {"command": ["elvish", "-lsp"], "languages": {"elvish"}},
    "erlang-ls": {"command": ["erlang_ls"], "languages": {"erlang"}},
    "forc": {"command": ["forc", "lsp"], "languages": {"sway"}},
    "forth-lsp": {"command": ["forth-lsp"], "languages": {"forth"}},
    "fortls": {
        "command": ["fortls", "--lowercase_intrinsics"],
        "languages": {"fortran"},
    },
    "fsharp-ls": {
        "config": {"AutomaticWorkspaceInit": True},
        "command": ["fsautocomplete"],
        "languages": {"fsharp"},
    },
    "gleam": {"command": ["gleam", "lsp"], "languages": {"gleam"}},
    "glsl_analyzer": {"command": ["glsl_analyzer"], "languages": {"glsl"}},
    "graphql-language-service": {
        "command": ["graphql-lsp", "server", "-m", "stream"],
        "languages": {"graphql"},
    },
    "haskell-language-server": {
        "command": ["haskell-language-server-wrapper", "--lsp"],
        "languages": {"cabal", "haskell"},
    },
    "hyprls": {"command": ["hyprls"], "languages": {"hyprlang"}},
    "idris2-lsp": {"command": ["idris2-lsp"], "languages": {"idris"}},
    "intelephense": {"command": ["intelephense", "--stdio"], "languages": {"php"}},
    "jdtls": {"command": ["jdtls"], "languages": {"java"}},
    "jedi": {"command": ["jedi-language-server"], "languages": {"python"}},
    "jq-lsp": {"command": ["jq-lsp"], "languages": {"jq"}},
    "jsonnet-language-server": {
        "command": ["jsonnet-language-server", "-t", "--lint"],
        "languages": {"jsonnet"},
    },
    "julia": {
        "timeout": 60,
        "command": [
            "julia",
            "--startup-file=no",
            "--history-file=no",
            "--quiet",
            "-e",
            "using LanguageServer; runserver()",
        ],
        "languages": {"julia"},
    },
    "koka": {
        "command": ["koka", "--language-server", "--lsstdio"],
        "languages": {"koka"},
    },
    "kotlin-language-server": {
        "command": ["kotlin-language-server"],
        "languages": {"kotlin"},
    },
    "lean": {"command": ["lean", "--server", "--memory=1024"], "languages": {"lean"}},
    "ltex-ls": {"command": ["ltex-ls"]},
    "markdoc-ls": {"command": ["markdoc-ls", "--stdio"], "languages": {"markdoc"}},
    "markdown-oxide": {"command": ["markdown-oxide"], "languages": {"markdown"}},
    "marksman": {"command": ["marksman", "server"], "languages": {"markdown"}},
    "metals": {
        "config": {
            "isHttpEnabled": True,
            "metals": {
                "inlayHints": {
                    "typeParameters": {"enable": True},
                    "hintsInPatternMatch": {"enable": True},
                }
            },
        },
        "command": ["metals"],
        "languages": {"scala"},
    },
    "mesonlsp": {"command": ["mesonlsp", "--lsp"], "languages": {"meson"}},
    "mint": {"command": ["mint", "ls"], "languages": {"mint"}},
    "mojo-lsp": {"command": ["mojo-lsp-server"], "languages": {"mojo"}},
    "nil": {"command": ["nil"], "languages": {"nix"}},
    "nimlangserver": {"command": ["nimlangserver"], "languages": {"nim"}},
    "nimlsp": {"command": ["nimlsp"]},
    "nixd": {"command": ["nixd"], "languages": {"nix"}},
    "nls": {"command": ["nls"], "languages": {"nickel"}},
    "nu-lsp": {"command": ["nu", "--lsp"], "languages": {"nu"}},
    "ocamllsp": {"command": ["ocamllsp"], "languages": {"ocaml", "ocaml-interface"}},
    "ols": {"command": ["ols"], "languages": {"odin"}},
    "omnisharp": {
        "command": ["OmniSharp", "--languageserver"],
        "languages": {"c-sharp"},
    },
    "openscad-lsp": {"command": ["openscad-lsp", "--stdio"], "languages": {"openscad"}},
    "pasls": {"command": ["pasls"], "languages": {"pascal"}},
    "pbkit": {"command": ["pb", "lsp"], "languages": {"protobuf"}},
    "perlnavigator": {"command": ["perlnavigator", "--stdio"], "languages": {"perl"}},
    "pest-language-server": {
        "command": ["pest-language-server"],
        "languages": {"pest"},
    },
    "prisma-language-server": {
        "command": ["prisma-language-server", "--stdio"],
        "languages": {"prisma"},
    },
    "purescript-language-server": {
        "command": ["purescript-language-server", "--stdio"],
        "languages": {"purescript"},
    },
    "pylsp": {"command": ["pylsp"], "languages": {"snakemake", "python"}},
    "pyright": {"config": {}, "command": ["pyright-langserver", "--stdio"]},
    "basedpyright": {"config": {}, "command": ["basedpyright-langserver", "--stdio"]},
    "pylyzer": {"command": ["pylyzer", "--server"]},
    "qmlls": {"command": ["qmlls"], "languages": {"qml"}},
    "quint-language-server": {
        "command": ["quint-language-server", "--stdio"],
        "languages": {"quint"},
    },
    "r": {
        "command": ["R", "--no-echo", "-e", "languageserver::run()"],
        "languages": {"rmarkdown", "r"},
    },
    "racket": {
        "command": ["racket", "-l", "racket-langserver"],
        "languages": {"racket"},
    },
    "regols": {"command": ["regols"], "languages": {"rego"}},
    "rescript-language-server": {
        "command": ["rescript-language-server", "--stdio"],
        "languages": {"rescript"},
    },
    "robotframework_ls": {"command": ["robotframework_ls"], "languages": {"robot"}},
    "ruff": {"command": ["ruff", "server"], "languages": {"python"}},
    "serve-d": {"command": ["serve-d"], "languages": {"d"}},
    "slint-lsp": {"command": ["slint-lsp"], "languages": {"slint"}},
    "solargraph": {"command": ["solargraph", "stdio"], "languages": {"ruby"}},
    "solc": {"command": ["solc", "--lsp"], "languages": {"solidity"}},
    "sourcekit-lsp": {"command": ["sourcekit-lsp"], "languages": {"swift"}},
    "spade-language-server": {
        "command": ["spade-language-server"],
        "languages": {"spade"},
    },
    "svlangserver": {"command": ["svlangserver"], "languages": {"verilog"}},
    "swipl": {
        "command": [
            "swipl",
            "-g",
            "use_module(library(lsp_server))",
            "-g",
            "lsp_server:main",
            "-t",
            "halt",
            "--",
            "stdio",
        ],
        "languages": {"prolog"},
    },
    "superhtml": {"command": ["superhtml", "lsp"], "languages": {"html"}},
    "tailwindcss-ls": {"command": ["tailwindcss-language-server", "--stdio"]},
    "taplo": {"command": ["taplo", "lsp", "stdio"], "languages": {"toml"}},
    "templ": {"command": ["templ", "lsp"], "languages": {"templ"}},
    "terraform-ls": {
        "command": ["terraform-ls", "serve"],
        "languages": {"tfvars", "hcl"},
    },
    "texlab": {"command": ["texlab"], "languages": {"latex", "bibtex"}},
    "typespec": {"command": ["tsp-server", "--stdio"], "languages": {"typespec"}},
    "vala-language-server": {
        "command": ["vala-language-server"],
        "languages": {"vala"},
    },
    "vale-ls": {"command": ["vale-ls"]},
    "vhdl_ls": {"command": ["vhdl_ls"], "languages": {"vhdl"}},
    "vlang-language-server": {"command": ["v-analyzer"], "languages": {"v"}},
    "vscode-css-language-server": {
        "config": {"provideFormatter": True, "css": {"validate": {"enable": True}}},
        "command": ["vscode-css-language-server", "--stdio"],
        "languages": {"scss", "css"},
    },
    "vscode-html-language-server": {
        "config": {"provideFormatter": True},
        "command": ["vscode-html-language-server", "--stdio"],
        "languages": {"html"},
    },
    "vscode-json-language-server": {
        "config": {"provideFormatter": True, "json": {"validate": {"enable": True}}},
        "command": ["vscode-json-language-server", "--stdio"],
        "languages": {"jsonc", "json"},
    },
    "vuels": {
        "config": {"typescript": {"tsdk": "node_modules/typescript/lib/"}},
        "command": ["vue-language-server", "--stdio"],
        "languages": {"vue"},
    },
    "wgsl_analyzer": {"command": ["wgsl_analyzer"], "languages": {"wgsl"}},
    "yaml-language-server": {
        "command": ["yaml-language-server", "--stdio"],
        "languages": {"yaml", "docker-compose"},
    },
    "zls": {"command": ["zls"], "languages": {"zig"}},
    "blueprint-compiler": {
        "command": ["blueprint-compiler", "lsp"],
        "languages": {"blueprint"},
    },
    "typst-lsp": {"command": ["typst-lsp"], "languages": {"typst"}},
    "tinymist": {"command": ["tinymist"], "languages": {"typst"}},
    "pkgbuild-language-server": {
        "command": ["pkgbuild-language-server"],
        "languages": {"pkgbuild"},
    },
    "helm_ls": {"command": ["helm_ls", "serve"], "languages": {"helm"}},
    "ember-language-server": {
        "command": ["ember-language-server", "--stdio"],
        "languages": {"gts", "gjs", "glimmer"},
    },
    "ansible-language-server": {
        "command": ["ansible-language-server", "--stdio"],
        "languages": {"yaml"},
    },
    "lua-language-server": {
        "config": {
            "Lua": {
                "hint": {
                    "enable": True,
                    "arrayIndex": "Enable",
                    "setType": True,
                    "paramName": "All",
                    "paramType": True,
                    "await": True,
                }
            }
        },
        "command": ["lua-language-server"],
        "languages": {"lua"},
    },
    "gopls": {
        "config": {
            "hints": {
                "assignVariableTypes": True,
                "compositeLiteralFields": True,
                "constantValues": True,
                "functionTypeParameters": True,
                "parameterNames": True,
                "rangeVariableTypes": True,
            }
        },
        "command": ["gopls"],
        "languages": {"go", "gowork", "gotmpl", "gomod"},
    },
    "golangci-lint-lsp": {
        "config": {
            "command": [
                "golangci-lint",
                "run",
                "--out-format",
                "json",
                "--issues-exit-code=1",
            ]
        },
        "command": ["golangci-lint-langserver"],
        "languages": {"go"},
    },
    "rust-analyzer": {
        "config": {
            "inlayHints": {
                "bindingModeHints": {"enable": False},
                "closingBraceHints": {"minLines": 10},
                "closureReturnTypeHints": {"enable": "with_block"},
                "discriminantHints": {"enable": "fieldless"},
                "lifetimeElisionHints": {"enable": "skip_trivial"},
                "typeHints": {"hideClosureInitialization": False},
            }
        },
        "command": ["rust-analyzer"],
        "languages": {"rust"},
    },
    "typescript-language-server": {
        "config": {
            "hostInfo": "helix",
            "typescript": {
                "inlayHints": {
                    "includeInlayEnumMemberValueHints": True,
                    "includeInlayFunctionLikeReturnTypeHints": True,
                    "includeInlayFunctionParameterTypeHints": True,
                    "includeInlayParameterNameHints": "all",
                    "includeInlayParameterNameHintsWhenArgumentMatchesName": True,
                    "includeInlayPropertyDeclarationTypeHints": True,
                    "includeInlayVariableTypeHints": True,
                }
            },
            "javascript": {
                "inlayHints": {
                    "includeInlayEnumMemberValueHints": True,
                    "includeInlayFunctionLikeReturnTypeHints": True,
                    "includeInlayFunctionParameterTypeHints": True,
                    "includeInlayParameterNameHints": "all",
                    "includeInlayParameterNameHintsWhenArgumentMatchesName": True,
                    "includeInlayPropertyDeclarationTypeHints": True,
                    "includeInlayVariableTypeHints": True,
                }
            },
        },
        "command": ["typescript-language-server", "--stdio"],
        "languages": {"jsx", "typescript", "javascript", "tsx"},
    },
    "svelteserver": {
        "config": {
            "configuration": {
                "typescript": {
                    "inlayHints": {
                        "parameterTypes": {"enabled": True},
                        "variableTypes": {"enabled": True},
                        "propertyDeclarationTypes": {"enabled": True},
                        "functionLikeReturnTypes": {"enabled": True},
                        "enumMemberValues": {"enabled": True},
                        "parameterNames": {"enabled": "all"},
                    }
                },
                "javascript": {
                    "inlayHints": {
                        "parameterTypes": {"enabled": True},
                        "variableTypes": {"enabled": True},
                        "propertyDeclarationTypes": {"enabled": True},
                        "functionLikeReturnTypes": {"enabled": True},
                        "enumMemberValues": {"enabled": True},
                        "parameterNames": {"enabled": "all"},
                    }
                },
            }
        },
        "command": ["svelteserver", "--stdio"],
        "languages": {"svelte"},
    },
    "vscode-eslint-language-server": {
        "config": {
            "validate": "on",
            "experimental": {"useFlatConfig": False},
            "rulesCustomizations": [],
            "run": "onType",
            "problems": {"shortenToSingleLine": False},
            "nodePath": "",
            "codeAction": {
                "disableRuleComment": {"enable": True, "location": "separateLine"},
                "showDocumentation": {"enable": True},
            },
            "workingDirectory": {"mode": "location"},
        },
        "command": ["vscode-eslint-language-server", "--stdio"],
        "languages": {"gts", "gjs"},
    },
}



"""Contain various application tab implementations."""

from __future__ import annotations

from dataclasses import dataclass, field
from pkgutil import resolve_name
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from euporie.core.tabs.base import Tab


@dataclass
class TabRegistryEntry:
    """Class to store tab information."""

    path: str
    name: str
    mime_types: set = field(default_factory=set)
    file_extensions: dict[str, None] = field(default_factory=dict)
    weight: int = 0

    @property
    def tab_class(self) -> type[Tab]:
        """Import and return the tab class."""
        return resolve_name(self.path)

    def __lt__(self, other: TabRegistryEntry) -> bool:
        """Sort by weight."""
        return self.weight < other.weight

    def __hash__(self) -> int:
        """Make the class hashable based on its path."""
        return hash(self.path)

    def __eq__(self, other: object) -> bool:
        """Compare TabRegistryEntry objects based on their path."""
        if not isinstance(other, TabRegistryEntry):
            return NotImplemented
        return self.path == other.path


_TAB_REGISTRY: list[TabRegistryEntry] = []



"""Module concerning clipboard access and manipulation."""

from __future__ import annotations

import logging

from prompt_toolkit.clipboard.base import Clipboard, ClipboardData
from prompt_toolkit.clipboard.in_memory import InMemoryClipboard
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit.selection import SelectionType

from euporie.core.app.current import get_app
from euporie.core.io import Vt100_Output
from euporie.core.key_binding.key_processor import KeyProcessor

log = logging.getLogger(__name__)


class Osc52Clipboard(Clipboard):
    """Clipboard that syncs with the system clipboard using OSC52 escape codes."""

    def __init__(self) -> None:
        """Create a new instance of the clipboard."""
        self._data = ClipboardData()

    def set_data(self, data: ClipboardData) -> None:
        """Set clipboard data."""
        output = get_app().output
        if isinstance(output, Vt100_Output):
            output.set_clipboard(data.text)
            output.flush()
        self._data = data

    def get_data(self) -> ClipboardData:
        """Retrieve clipboard data."""
        # Send clipboard query
        app = get_app()
        output = app.output
        # Request clipboard contents from terminal
        if isinstance(output, Vt100_Output):
            from euporie.core.keys import MoreKeys

            output.get_clipboard()
            output.flush()
            # Wait for terminal response
            if isinstance(app.key_processor, KeyProcessor):
                app.key_processor.await_key(MoreKeys.ClipboardDataResponse)
        return self._data

    def sync(self, text: str) -> None:
        """Update the last known clipboard data."""
        if text != self._data.text:
            self._data = ClipboardData(
                text=text,
                type=SelectionType.LINES if "\n" in text else SelectionType.CHARACTERS,
            )


CONFIGURED_CLIPBOARDS = {
    "internal": InMemoryClipboard,
    "external": PyperclipClipboard,
    "terminal": Osc52Clipboard,
}



"""Contain functions which can be used to convert data to multiple formats."""

from __future__ import annotations

import base64
import logging
from typing import TYPE_CHECKING

from euporie.core.app.current import get_app
from euporie.core.convert.utils import call_subproc, scale_to_fit

if TYPE_CHECKING:
    from typing import Any, Literal

    from euporie.core.convert.datum import Datum

log = logging.getLogger(__name__)


async def base64_to_bytes_py(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> bytes:
    """Convert base64 encoded data to bytes."""
    data = datum.data
    data_str = data.decode() if isinstance(data, bytes) else data
    return base64.b64decode(data_str)


async def imagemagick_convert(
    output_format: str,
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str | bytes:
    """Convert image data to PNG bytes using ``imagemagick``."""
    cmd: list[Any] = ["magick"]
    if bg:
        cmd += ["-background", bg]
    cmd.extend(["-[0]", "-density", "300"])
    app = get_app()
    if cols is not None and hasattr(app, "cell_size_px"):
        px, _ = app.cell_size_px
        cmd += ["-geometry", f"{int(cols * px)}"]
    cmd += [f"{output_format}:-"]
    result: bytes | str = await call_subproc(datum.data, cmd)

    if output_format in {"sixel", "svg"} and isinstance(result, bytes):
        result = result.decode()
    return result


async def chafa_convert_cmd(
    output_format: str,
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str | bytes:
    """Convert image data to ANSI text using :command:`chafa`."""
    cmd: list[Any] = ["chafa", f"--format={output_format}"]
    if cols is not None or rows is not None:
        size = "--size="
        if cols is not None:
            size = f"{size}{cols}"
        if rows is not None:
            size = f"{size}x{rows}"
        cmd.append(size)
    if bg:
        cmd += ["--bg", bg]
    # cmd += ["--stretch", "/dev/stdin"]
    return (await call_subproc(datum.data, cmd)).decode()


async def chafa_convert_py(
    output_format: Literal["symbols", "sixels", "kitty", "iterm2"],
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str | bytes:
    """Convert image data to ANSI text using ::`chafa.py`."""
    from chafa import Canvas, CanvasConfig, PixelMode, PixelType
    from PIL import Image

    pil_mode_to_pixel_type = {
        "RGBa": PixelType.CHAFA_PIXEL_RGBA8_PREMULTIPLIED,
        "RGBA": PixelType.CHAFA_PIXEL_RGBA8_UNASSOCIATED,
        "RGB": PixelType.CHAFA_PIXEL_RGB8,
    }

    str_to_pixel_mode = {
        "symbols": PixelMode.CHAFA_PIXEL_MODE_SYMBOLS,
        "sixels": PixelMode.CHAFA_PIXEL_MODE_SIXELS,
        "kitty": PixelMode.CHAFA_PIXEL_MODE_KITTY,
        "iterm2": PixelMode.CHAFA_PIXEL_MODE_ITERM2,
    }

    # Convert PIL image to format that chafa can use
    data = datum.data
    # Always convert the image, as unconverted images sometime result in an off-by-one
    # line width errors resulting in diagonal image striping for some reason
    data = data.convert("RGBA", palette=Image.Palette.ADAPTIVE, colors=16)
    # Init canvas config
    config = CanvasConfig()
    # Set output mode
    config.pixel_mode = str_to_pixel_mode[output_format]
    # Configure the canvas geometry based on our cell size
    config.width, config.height = await scale_to_fit(datum, cols, rows)

    # Set the foreground color
    if fg and (color := fg.lstrip("#")):
        config.fg_color = (
            int(color[0:2], 16),
            int(color[2:4], 16),
            int(color[4:6], 16),
        )

    # Set the background color
    if bg and (color := bg.lstrip("#")):
        config.bg_color = (
            int(color[0:2], 16),
            int(color[2:4], 16),
            int(color[4:6], 16),
        )

    # Init the canvas
    canvas = Canvas(config)

    # Draw to canvas
    canvas.draw_all_pixels(
        pil_mode_to_pixel_type.get(data.mode, PixelType.CHAFA_PIXEL_RGBA8_UNASSOCIATED),
        list(data.tobytes()),
        width := data.width,
        data.height,
        width * len(data.getbands()),
    )

    # Return the output
    return canvas.print(fallback=True).decode()



"""Concerns the interaction with kernels."""

from __future__ import annotations

from importlib.util import find_spec
from pkgutil import resolve_name
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Literal

    from euporie.core.kernel.base import BaseKernel, KernelInfo, MsgCallbacks
    from euporie.core.tabs.kernel import KernelTab

KERNEL_REGISTRY = {
    "local": "euporie.core.kernel.local:LocalPythonKernel",
}
if find_spec("jupyter_client"):
    KERNEL_REGISTRY["jupyter"] = "euporie.core.kernel.jupyter:JupyterKernel"


def list_kernels() -> list[KernelInfo]:
    """Get specifications for all available kernel types.

    Returns:
        A dictionary mapping kernel type names to their specifications.
    """
    return [
        variant
        for type_path in KERNEL_REGISTRY.values()
        for variant in resolve_name(type_path).variants()
    ]


def create_kernel(
    type_name: Literal["jupyter", "local"],
    kernel_tab: KernelTab,
    default_callbacks: MsgCallbacks | None = None,
    allow_stdin: bool = False,
    **kwargs: Any,
) -> BaseKernel:
    """Create and return appropriate kernel instance."""
    type_path = KERNEL_REGISTRY.get(type_name)
    if type_path is not None:
        type_class = resolve_name(type_path)
        return type_class(
            kernel_tab=kernel_tab,
            default_callbacks=default_callbacks,
            allow_stdin=allow_stdin,
            **kwargs,
        )
    else:
        raise ValueError(f"Unknown kernel type: {type_name}")
